#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Register or fetch Pipes from the API
"""

from __future__ import annotations
from meerschaum.utils.typing import SuccessTuple, Union, Any, Optional, Mapping, List

def pipe_r_url(
        pipe : 'meerschaum.Pipe'
    ) -> str:
    """
    Generate a relative URL path from a Pipe's keys.
    """
    from meerschaum.config.static import _static_config
    location_key = pipe.location_key
    if location_key is None:
        location_key = '[None]'
    return f"{_static_config()['api']['endpoints']['pipes']}/{pipe.connector_keys}/{pipe.metric_key}/{location_key}"

def register_pipe(
        self,
        pipe : meerschaum.Pipe,
        debug : bool = False
    ) -> SuccessTuple:
    """
    Submit a POST to the API to register a new Pipe object.
    Returns a tuple of (success_bool, response_dict)
    """
    from meerschaum.utils.debug import dprint
    from meerschaum.config.static import _static_config
    ### NOTE: if `parameters` is supplied in the Pipe constructor,
    ###       then `pipe.parameters` will exist and not be fetched from the database.
    response = self.post(_static_config()['api']['endpoints']['pipes'], json=pipe.meta, debug=debug)
    if debug:
        dprint(response.text)
    if isinstance(response.json(), list):
        response_tuple = response.__bool__(), response.json()[1]
    elif 'detail' in response.json():
        response_tuple = response.__bool__(), response.json()['detail']
    else:
        response_tuple = response.__bool__(), response.text
    return response_tuple

def edit_pipe(
        self,
        pipe : meerschaum.Pipe,
        patch : bool = False,
        debug : bool = False,
        **kw : Any
    ) -> SuccessTuple:
    """
    Submit a PATCH to the API to edit an existing Pipe.
    Returns a tuple of (success_bool, response_dict)
    """
    from meerschaum.config.static import _static_config
    if debug:
        from meerschaum.utils.debug import dprint
        dprint(f"patch: {patch}")
    response = self.patch(
        _static_config()['api']['endpoints']['pipes'],
        json = pipe.meta,
        params = {'patch' : patch},
        debug = debug
    )
    try:
        j = response.json()
    except Exception as e:
        return False, str(e)

    if isinstance(j, dict) and 'detail' in j:
        return False, j['detail']

    return tuple(j)

def fetch_pipes_keys(
        self,
        connector_keys : list = [],
        metric_keys : list = [],
        location_keys : list = [],
        params : dict = dict(),
        mrsm_instance : str = 'api',
        debug : bool = False
    ) -> Union[List[str], Mapping[str, Any]]:
    """
    NOTE: This function no longer builds Pipes. Use the main `get_pipes()` function
          with the arguments `mrsm_instance = 'api' and `method = 'registered'` (default).

    Fetch registered Pipes' keys from the API.

    keys_only : bool : True
        If True, only return a list of tuples of the keys
        E.g. [ (connector_keys, metric_key, location_key) ]
        This is used by the main `get_pipes()` function for the 'api' method.
    """
    from meerschaum.utils.warnings import error
    from meerschaum.config.static import _static_config
    import json

    r_url = _static_config()['api']['endpoints']['pipes'] + '/keys'
    try:
        j = self.get(
            r_url,
            params = {
                'connector_keys' : json.dumps(connector_keys),
                'metric_keys' : json.dumps(metric_keys),
                'location_keys' : json.dumps(location_keys),
                'params' : json.dumps(params),
                'debug' : debug,
            },
            debug=debug
        ).json()
    except Exception as e:
        error(str(e))

    if 'detail' in j:
        error(j['detail'], stack=False)
    return [tuple(r) for r in j]

def sync_pipe(
        self,
        pipe : Optional[meerschaum.Pipe] = None,
        df : Optional[Union[pandas.DataFrame, Dict[Any, Any], str]] = None,
        chunksize : Optional[int] = -1,
        debug : bool = False,
        **kw : Any
    ) -> SuccessTuple:
    """
    Append a pandas DataFrame to a Pipe.
    If Pipe does not exist, it is registered with supplied metadata
        NOTE: columns['datetime'] must be set for new Pipes.
    """
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.misc import json_serialize_datetime
    from meerschaum.config import get_config
    from meerschaum.utils.packages import attempt_import
    import json, time
    begin = time.time()
    more_itertools = attempt_import('more_itertools')
    if df is None:
        msg = f"DataFrame is None. Cannot sync pipe '{pipe}"
        return False, msg

    def get_json_str(c):
        ### allow syncing dict or JSON without needing to import pandas (for IOT devices)
        if isinstance(c, dict):
            json_str = json.dumps(c, default=json_serialize_datetime)
        else:
            json_str = c.to_json(date_format='iso', date_unit='us')
        return json_str

    df = json.loads(df) if isinstance(df, str) else df

    ### TODO Make separate chunksize for API?
    _chunksize : Optional[int] = (1 if chunksize is None else (
        get_config('system', 'connectors', 'sql', 'chunksize') if chunksize == -1
        else chunksize
    ))
    keys : list = list(df.keys())
    chunks = []
    if hasattr(df, 'index'):
        rowcount = len(df)
        chunks = [df.iloc[i] for i in more_itertools.chunked(df.index, _chunksize)]
    elif isinstance(df, dict):
        ### `_chunks` is a dict of lists of dicts.
        ### e.g. {'a' : [ {'a':[1, 2]}, {'a':[3, 4]} ] }
        _chunks = {k:[] for k in keys}
        rowcount = len(df[keys[0]])
        for k in keys:
            if len(df[k]) != rowcount:
                return False, "Arrays must all be the same length."
            chunk_iter = more_itertools.chunked(df[k], _chunksize)
            for l in chunk_iter:
                _chunks[k].append({k:l})

        ### `chunks` is a list of dicts (e.g. orient by rows in pandas JSON).
        for k, l in _chunks.items():
            for i, c in enumerate(l):
                try:
                    chunks[i].update(c)
                except IndexError:
                    chunks.append(c)

    ### Send columns in case the user has defined them locally.
    if pipe.columns:
        kw['columns'] = json.dumps(pipe.columns)
    r_url = pipe_r_url(pipe) + '/data'

    for i, c in enumerate(chunks):
        if debug:
            dprint(f"Posting chunk ({i + 1} / {_chunksize}) to {r_url}...")
            print(c)
        json_str = get_json_str(c)

        try:
            response = self.post(
                r_url,
                ### handles check_existing
                params = kw,
                data = json_str,
                debug = debug
            )
        except Exception as e:
            warn(str(e))
            return False, str(e)
            
        if not response:
            return False, f"Failed to receive response. Response text: {response.text}"

        try:
            j = json.loads(response.text)
        except Exception as e:
            return False, str(e)

        if isinstance(j, dict) and 'detail' in j:
            return False, j['detail']

        try:
            j = tuple(j)
        except Exception as e:
            return False, response.text

        if debug:
            dprint("Received response: " + str(j))
        if not j[0]:
            return j

    len_chunks = len(chunks)
    return True, (
        f"It took {round(time.time() - begin, 2)} seconds to sync {rowcount} row" +
        ('s' if rowcount != 1 else '') + f" across {len_chunks} chunk" + ('s' if len_chunks != 1 else '') +
        f" to '{pipe}'."
    )

def delete_pipe(
        self,
        pipe : Optional[meerscahum.Pipe] = None,
        debug : bool = None,        
    ) -> SuccessTuple:
    """
    Delete a Pipe and drop its table.
    """
    from meerschaum.utils.warnings import error
    if pipe is None:
        error(f"Pipe cannot be None.")
    return self.do_action(
        ['delete', 'pipes'],
        connector_keys = pipe.connector_keys,
        metric_keys = pipe.metric_key,
        location_keys = pipe.location_key,
        force = True,
        debug = debug
    )

def get_pipe_data(
        self,
        pipe : meerschaum.Pipe,
        begin : Optional[datetime.datetime] = None,
        end : Optional[datetime.datetime] = None,
        params : Optional[Dict[str, Any]] = None,
        as_chunks : bool = False,
        debug : bool = False,
        **kw: Any
    ) -> Optional[pandas.DataFrame]:
    """
    Fetch data from the API.
    """
    from meerschaum.utils.warnings import warn
    r_url = pipe_r_url(pipe)
    chunks_list = []
    while True:
        try:
            response = self.get(
                r_url + "/data",
                json = params,
                params = {'begin': begin, 'end': end},
                debug = debug
            )
            j = response.json()
        except Exception as e:
            warn(str(e))
            return None
        if isinstance(j, dict) and 'detail' in j:
            return False, j['detail']
        break
    from meerschaum.utils.packages import import_pandas
    from meerschaum.utils.misc import parse_df_datetimes
    pd = import_pandas()
    try:
        df = pd.read_json(response.text)
    except Exception as e:
        warn(str(e))
        return None
    df = parse_df_datetimes(pd.read_json(response.text), debug=debug)
    return df

def get_backtrack_data(
        self,
        pipe : meerschaum.Pipe,
        begin : datetime.datetime,
        backtrack_minutes : int = 0,
        debug : bool = False,
        **kw : Any,
    ) -> pandas.DataFrame:
    """
    Get a Pipe's backtrack data from the API.
    """
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.warnings import warn
    r_url = pipe_r_url(pipe)
    try:
        response = self.get(
            r_url + "/backtrack_data",
            params = {
                'begin': begin,
                'backtrack_minutes': backtrack_minutes,
            },
            debug = debug
        )
    except Exception as e:
        warn(f"Failed to parse backtrack data JSON for pipe '{pipe}'. Exception:\n" + str(e))
        return None
    from meerschaum.utils.packages import import_pandas
    from meerschaum.utils.misc import parse_df_datetimes
    if debug:
        dprint(response.text)
    pd = import_pandas()
    try:
        df = pd.read_json(response.text)
    except Exception as e:
        warn(str(e))
        return None
    df = parse_df_datetimes(pd.read_json(response.text), debug=debug)
    return df

def get_pipe_id(
        self,
        pipe : meerschuam.Pipe,
        debug : bool = False,
    ) -> int:
    """
    Get a Pipe's ID from the API
    """
    from meerschaum.utils.debug import dprint
    r_url = pipe_r_url(pipe)
    response = self.get(
        r_url + '/id',
        debug = debug
    )
    if debug:
        dprint(response.text)
    try:
        return int(response.text)
    except Exception as e:
        return None

def get_pipe_attributes(
        self,
        pipe : meerschaum.Pipe,
        debug : bool = False,
    ) -> Mapping[str, Any]:
    """
    Get a Pipe's attributes from the API
    """
    r_url = pipe_r_url(pipe)
    response = self.get(r_url + '/attributes', debug=debug)
    import json
    try:
        return json.loads(response.text)
    except Exception as e:
        return None

def get_sync_time(
        self,
        pipe : 'meerschaum.Pipe',
        params : dict = None,
        debug : bool = False,
    ) -> datetime.datetime:
    """
    Get a Pipe's most recent datetime value from the API
    """
    import datetime, json
    r_url = pipe_r_url(pipe)
    response = self.get(r_url + '/sync_time', json=params, params={'debug' : debug}, debug=debug)
    if not response:
        return None
    j = response.json()
    if j is None:
        dt = None
    else:
        try:
            dt = datetime.datetime.fromisoformat(json.loads(response.text))
        except Exception as e:
            dt = None
    return dt

def pipe_exists(
        self,
        pipe : 'meerschaum.Pipe',
        debug : bool = False
    ) -> bool:
    """
    Consult the API to see if a Pipe exists
    """
    r_url = pipe_r_url(pipe)
    response = self.get(r_url + '/exists', debug=debug)
    if debug:
        dprint("Received response: " + str(response.text))
    j = response.json()
    if isinstance(j, dict) and 'detail' in j:
        return False, j['detail']
    return j

def create_metadata(
        self,
        debug : bool = False
    ) -> bool:
    """
    Create Pipe metadata tables
    """
    from meerschaum.config.static import _static_config
    import json
    r_url = _static_config()['api']['endpoints']['metadata']
    response = self.post(r_url, debug=debug)
    if debug:
        dprint("Create metadata response: {response.text}")
    try:
        metadata_response = json.loads(response.text)
    except Exception as e:
        metadata_response = False
    return False

def get_pipe_rowcount(
        self,
        pipe : 'meerschaum.Pipe',
        begin : 'datetime.datetime' = None,
        end : 'datetime.datetime' = None,
        params : Optional[Dict[str, Any]] = None,
        remote : bool = False,
        debug : bool = False,
    ) -> Optional[int]:
    """
    Get a pipe's row couunt from the API.
    """
    import json
    r_url = pipe_r_url(pipe)
    response = self.get(
        r_url + "/rowcount",
        json = params,
        params = {
            'begin' : begin,
            'end' : end,
            'remote' : remote,
        },
        debug = debug
    )
    try:
        return int(json.loads(response.text))
    except Exception as e:
        return None

def drop_pipe(
        self,
        pipe : meerschaum.Pipe,
        debug : bool = False
    ) -> SuccessTuple:
    """
    Drop a pipe's tables but maintain its registration.
    """
    return self.do_action(
        ['drop', 'pipes'],
        connector_keys = pipe.connector_keys,
        metric_keys = pipe.metric_key,
        location_keys = pipe.location_key,
        force = True,
        debug = debug
    )

def get_pipe_columns_types(
        self,
        pipe : meerschaum.Pipe,
        debug : bool = False,
    ) -> Optional[Dict[str, str]]:
    """
    Return a dictionary of a pipe's table's columns to their data types.

    E.g. An example dictionary for a small table.

    ```
    >>> {
    ...   'dt': 'TIMESTAMP WITHOUT TIMEZONE',
    ...   'id': 'BIGINT',
    ...   'val': 'DOUBLE PRECISION',
    ... }
    >>> 
    ```
    """
    r_url = pipe_r_url(pipe) + '/columns/types'
    response = self.get(
        r_url,
        debug = debug
    )
    j = response.json()
    if isinstance(j, dict) and 'detail' in j and len(j.keys()) == 1:
        from meerschaum.utils.warnings import warn
        warn(j['detail'])
        return None
    elif not isinstance(j, dict):
        warn(response.text)
        return None
    return j



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
This module contains the logic that builds the sqlalchemy engine string.
"""

from meerschaum.utils.debug import dprint
from meerschaum.config._paths import SQLITE_DB_PATH, DUCKDB_PATH

### determine driver and requirements from flavor
default_requirements = {
    'username',
    'password',
    'host',
    'database',
}
default_create_engine_args = {
    #  'method',
    'pool_size',
    'max_overflow',
    'pool_recycle',
    'connect_args',
}
flavor_configs = {
    'timescaledb'      : {
        'engine'       : 'postgresql',
        'create_engine' : default_create_engine_args,
        'requirements' : default_requirements,
        'defaults'     : {
            'port'     : 5432,
        },
    },
    'postgresql'         : {
        'engine'       : 'postgresql',
        'create_engine' : default_create_engine_args,
        'requirements' : default_requirements,
        'defaults'     : {
            'port'     : 5432,
        },
    },
    'mssql'            : {
        'engine'       : 'mssql+pyodbc',
        'create_engine' : default_create_engine_args,
        'requirements' : default_requirements,
        'defaults'     : {
            'port'     : 1433,
            'driver'   : 'ODBC+Driver+17+for+SQL+Server',
        },
    },
    'mysql'            : {
        'engine'       : 'mysql+pymysql',
        'create_engine' : default_create_engine_args,
        'requirements' : default_requirements,
        'defaults'     : {
            'port'     : 3306,
        },
    },
    'mariadb'          : {
        'engine'       : 'mysql+pymysql',
        'create_engine' : default_create_engine_args,
        'requirements' : default_requirements,
        'defaults'     : {
            'port'     : 3306,
        },
    },
    'oracle'           : {
        'engine'       : 'oracle+cx_oracle',
        'create_engine' : default_create_engine_args,
        'requirements' : default_requirements,
        'defaults'     : {
            'port'     : 1521,
        },
    },
    'sqlite'           : {
        'engine'       : 'sqlite',
        'create_engine' : default_create_engine_args,
        'requirements' : {
        },
        'defaults'     : {
            'database' : SQLITE_DB_PATH,
        },
    },
    'duckdb' : {
        'engine' : 'duckdb',
        'create_engine' : {},
        'requirements' : '',
        'defaults' : {
            'database' : DUCKDB_PATH,
        },
    },
    'cockroachdb'      : {
        'engine'       : 'cockroachdb',
        'create_engine' : {c for c in default_create_engine_args if c != 'method'},
        'requirements' : {'host'},
        'defaults'     : {
            'port'     : 26257,
            'database' : 'defaultdb',
            'username' : 'root',
            'password' : 'admin',
        },
    },
}
install_flavor_drivers = {
    'sqlite' : ['aiosqlite'],
    'duckdb' : ['duckdb', 'duckdb_engine'],
    'mysql' : ['pymysql'],
    'timescaledb' : ['psycopg2'],
    'postgresql' : ['psycopg2'],
    'cockroachdb' : ['psycopg2', 'cockroachdb'],
}


def create_engine(
        self,
        include_uri : bool = False,
        debug : bool = False,
        **kw
    ) -> 'sqlalchemy.engine.Engine':
    """
    Create a sqlalchemy engine by building the engine string.

    returns: sqlalchemy engine
    """
    from meerschaum.utils.packages import attempt_import
    from meerschaum.utils.warnings import error, warn
    sqlalchemy = attempt_import('sqlalchemy')
    import urllib
    if self.flavor in install_flavor_drivers:
        attempt_import(*install_flavor_drivers[self.flavor], debug=self._debug, lazy=False)

    ### supplement missing values with defaults (e.g. port number)
    for a, value in flavor_configs[self.flavor]['defaults'].items():
        if a not in self.__dict__:
            self.__dict__[a] = value

    ### Verify that everything is in order.
    if self.flavor not in flavor_configs:
        error(f"Cannot create a connector with the flavor '{self.flavor}'.")

    _engine = flavor_configs[self.flavor].get('engine', None)
    _username = self.__dict__.get('username', None)
    _password = self.__dict__.get('password', None)
    _host = self.__dict__.get('host', None)
    _port = self.__dict__.get('port', None)
    _database = self.__dict__.get('database', None)
    _driver = self.__dict__.get('driver', None)

    ### self.sys_config was deepcopied and can be updated safely
    if self.flavor in ("sqlite", "duckdb"):
        ### The duckdb dialect might not be registered.
        if self.flavor == 'duckdb':
            sqlalchemy.engine.url.registry.register("duckdb", "duckdb_engine", "Dialect")

        engine_str = f"{_engine}:///{_database}"
        if 'create_engine' not in self.sys_config:
            self.sys_config['create_engine'] = {}
        if 'connect_args' not in self.sys_config['create_engine']:
            self.sys_config['create_engine']['connect_args'] = {}
        self.sys_config['create_engine']['connect_args'].update({"check_same_thread" : False})
    else:
        engine_str = (
            _engine + "://" + (_username if _username is not None else '') +
            ((":" + urllib.parse.quote_plus(_password)) if _password is not None else '') +
            "@" + _host + ((":" + str(_port)) if _port is not None else '') +
            (("/" + _database) if _database is not None else '')
            + (("?driver=" + _driver) if _driver is not None else '')
        )
    if debug:
        dprint(
            (
                (engine_str.replace(':' + _password, ':' + ('*' * len(_password))))
                    if _password is not None else engine_str
            ) + '\n' + f"{self.sys_config.get('create_engine', {}).get('connect_args', {})}"
        )

    _kw_copy = kw.copy()
    _create_engine_args = {
        k: v for k, v in self.sys_config.get('create_engine', {}).items()
            if k in flavor_configs[self.flavor].get('create_engine', {})
    }
    _create_engine_args.update(_kw_copy)

    try:
        engine = sqlalchemy.create_engine(
            engine_str,
            ### I know this looks confusing, and maybe it's bad code,
            ### but it's simple. It dynamically parses the config string
            ### and splits it to separate the class name (QueuePool)
            ### from the module name (sqlalchemy.pool).
            poolclass    = getattr(
                attempt_import(
                    ".".join(self.sys_config['poolclass'].split('.')[:-1])
                ),
                self.sys_config['poolclass'].split('.')[-1]
            ),
            echo         = debug,
            **_create_engine_args
        )
    except Exception as e:
        warn(e)
        warn(f"Failed to create connector '{self}'.", stack=False)
        engine = None

    if include_uri:
        return engine, engine_str
    return engine



#, manager! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Routes for managing users
"""

from __future__ import annotations
from meerschaum.utils.typing import (
    Optional, Union, SuccessTuple, Any, Mapping, Sequence, Dict, List
)

from meerschaum.utils.packages import attempt_import
from meerschaum.api import (
    fastapi, app, endpoints, get_api_connector, pipes, get_pipe,
    manager, debug, check_allow_chaining, DISALLOW_CHAINING_MESSAGE
)
from meerschaum.api.tables import get_tables
from starlette.responses import Response, JSONResponse
from meerschaum._internal.User import User
import os, pathlib, datetime

import meerschaum._internal.User
sqlalchemy = attempt_import('sqlalchemy')
users_endpoint = endpoints['users']

from fastapi import HTTPException
from meerschaum.config.static import _static_config

@app.get(users_endpoint + "/me")
def read_current_user(
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
    ) -> Dict[str, Union[str, int]]:
    """
    Return attributes of the current User.
    """
    return {
        'username' : curr_user.username,
        'user_id' : get_api_connector().get_user_id(curr_user),
        'user_type' : get_api_connector().get_user_type(curr_user),
        'attributes' : get_api_connector().get_user_attributes(curr_user),
    }

@app.get(users_endpoint)
def get_users() -> List[str]:
    """
    Return a list of registered users
    """
    return get_api_connector().get_users(debug=debug)

@app.post(users_endpoint + "/{username}/register")
def register_user(
        username : str,
        password : str,
        type : Optional[str] = None,
        email : Optional[str] = None,
        attributes : Optional[Dict[str, Any]] = None
    ) -> SuccessTuple:
    """
    Register a new user.
    """
    from meerschaum.config import get_config
    allow_users = get_config('system', 'api', 'permissions', 'registration', 'users', patch=True)
    if not allow_users:
        return False, (
            "The administrator for this server has not allowed user registration.\n\n" +
            "Please contact the system administrator, or if you are running this server, " +
            "open the configuration file with `edit config system` and search for 'permissions'. " +
            " Under the keys api:permissions:registration, " +
            "you can toggle various registration types."
        )
    if type == 'admin':
        return False, (
            "New users cannot be of type 'admin' when using the API connector. " +
            "Register a normal user first, then edit the user from an authorized account, " +
            "or use a SQL connector instead."
        )
    user = User(username, password, type=type, email=email, attributes=attributes)
    return get_api_connector().register_user(user, debug=debug)

@app.post(users_endpoint + "/{username}/edit")
def edit_user(
        username : str,
        password : str,
        type : Optional[str] = None,
        email : Optional[str] = None,
        attributes : Optional[Dict[str, Any]] = None,
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
    ) -> SuccessTuple:
    """
    Edit an existing user
    """
    user = User(username, password, email=email, attributes=attributes)
    user_type = get_api_connector().get_user_type(curr_user)
    if user_type == 'admin' and type is not None:
        user.type = type
    if user_type == 'admin' or curr_user.username == user.username:
        return get_api_connector().edit_user(user, debug=debug)

    return False, f"Cannot edit user '{user}': Permission denied"

@app.get(users_endpoint + "/{username}/id")
def get_user_id(
        username : str,
    ) -> Optional[int]:
    """
    Get a user's ID
    """
    return get_api_connector().get_user_id(User(username), debug=debug)

@app.get(users_endpoint + "/{username}/attributes")
def get_user_attributes(
        username : str,
    ) -> Optional[Dict[str, Any]]:
    """
    Get a user's attributes.
    """
    return get_api_connector().get_user_attributes(User(username), debug=debug)

@app.delete(users_endpoint + "/{username}")
def delete_user(
        username : str,
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
    ) -> SuccessTuple:
    """
    Delete a user.
    """
    user = User(username)
    user_type = get_api_connector().get_user_type(curr_user, debug=debug)
    if user_type == 'admin' or curr_user.username == user.username:
        return get_api_connector().delete_user(user, debug=debug)

    return False, f"Cannot delete user '{user}': Permission denied"

###################################
# Internal API Chaining functions #
###################################

@app.get(users_endpoint + '/{username}/password_hash')
def get_user_password_hash(
        username : str,
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
    ) -> Union[str, HTTPException]:
    """
    If configured to allow chaining, return a user's password_hash.
    """
    if not check_allow_chaining():
        raise HTTPException(status_code=403, detail=DISALLOW_CHAINING_MESSAGE)
    return get_api_connector().get_user_password_hash(User(username), debug=debug)

@app.get(users_endpoint + '/{username}/type')
def get_user_type(
        username : str,
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
    ) -> Union[str, HTTPException]:
    """
    If configured to allow chaining, return a user's type.
    """
    if not check_allow_chaining():
        raise HTTPException(status_code=403, detail=DISALLOW_CHAINING_MESSAGE)
    return get_api_connector().get_user_type(User(username))



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
The main dashboard layout.
"""

import uuid
from meerschaum.config import __doc__ as doc, get_config
from meerschaum.utils.misc import get_connector_labels
from meerschaum.utils.packages import attempt_import
from meerschaum.api import endpoints
dex = attempt_import('dash_extensions', lazy=False)
enrich = attempt_import('dash_extensions.enrich', lazy=False)
dbc = attempt_import('dash_bootstrap_components', lazy=False)
dcc = attempt_import('dash_core_components', warn=False)
html = attempt_import('dash_html_components', warn=False)
px = attempt_import('plotly.express', warn=False)
daq = attempt_import('dash_daq', warn=False)

from meerschaum.api.dash.components import (
    go_button, search_parameters_editor, keyboard, websocket, test_button,
    get_items_menu, bottom_buttons_content, console_div
)
from meerschaum.api.dash.keys import (
    keys_lists_content, text_tab_content, dropdown_tab_content
)

layout = html.Div(
    id = 'main-div',
    children = [
        websocket,
        keys_lists_content,
        keyboard,
        dcc.Interval(
            id = 'check-input-interval',
            interval = 1 * 1000,
            n_intervals = 0,
            disabled = True,
        ),
        dbc.Navbar(
            children = [
                html.A(
                    dbc.Row(
                        [
                            dbc.Col(html.Img(
                                src = endpoints['dash'] + "/assets/logo_48x48.png",
                                style = {'padding' : '0.5em'}
                            )),
                            dbc.Col(dbc.NavbarBrand(
                                "Meerschaum Web Interface",
                                className = "m1-2"
                            )),
                        ],
                        align = 'center',
                        no_gutters = True
                    ),
                    href = '#',
                ),
                dbc.NavbarToggler(id="navbar-toggler"),
                dbc.Collapse(
                    dbc.Row(
                        [
                            dbc.Col(
                                html.Div(
                                    className = 'dbc_dark',
                                    children = [
                                        dbc.Select(
                                            id = 'instance-select',
                                            bs_size = 'sm',
                                            options = [
                                                {'label' : i, 'value' : i}
                                                for i in get_connector_labels('sql', 'api')
                                            ],
                                            className = 'dbc_dark',
                                        )
                                    ]
                                ),
                                #  width = {'size' : 2},
                            ), ### end of instance column
                            dbc.Col(
                                html.Pre(html.A(doc, href='/docs')),
                                #  width = {'order' : 'last'},
                                style = {'padding-left' : '15px', 'margin-top' : '15px'},
                            ),
                        ],
                        no_gutters = True,
                        className = "ml-auto flex-nowrap mt-3 mt-md-0",
                        align = 'center',
                        #  className='navbar-nav ml-auto'
                    ),
                    id = 'navbar-collapse',
                    navbar = True,
                ), ### end of navbar-collapse
            ],
            color = 'dark',
            dark = True
        ), ### end nav-bar
        dbc.Row(
            id = 'content-row',
            children = [
                dbc.Col(
                    children = [
                        dbc.Tabs(
                            id = 'pipes-filter-tabs',
                            children = [
                                dbc.Tab(
                                    dropdown_tab_content, label='Filter',
                                    id='pipes-filter-dropdown-tab',
                                    tab_id='dropdown',
                                ),
                                dbc.Tab(
                                    text_tab_content,
                                    label='Text',
                                    id='pipes-filter-input-tab',
                                    tab_id='input',
                                    tab_style={"display": "none"},
                                ),
                            ]
                        ),
                        html.Br(),
                        bottom_buttons_content,
                        #  action_row,
                        test_button,
                        #  go_button,
                        #  show_pipes_button,
                        #  get_items_menu,
                        html.Div(id='ws-div'),
                        #  search_parameters_editor,
                    ],
                    id = 'content-col-left',
                    width = {'size' : 6},
                ),
                dbc.Col(
                    children = [
                        dbc.Col([
                            ### Place alert divs here.
                                html.Div(id='success-alert-div'),
                                html.Div(id='instance-alert-div')
                            ],
                            width={'size' : 8, 'offset' : 2}
                        ),
                        html.Div(
                            id = 'content-div-right',
                            children = [console_div],
                        )
                    ],
                    width = {'size' : 6},
                    id = 'content-col-right',
                ),
            ],
            style = {'max-width' : '100%', 'padding' : '15px'},
        ),
        #  dcc.DatePickerRange(
            #  id='begin_end',
        #  ),
    ],
)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Define default values for the formatting key.
"""

default_unicode, default_ansi = True, True
import platform
if platform.system() == 'Windows':
    default_unicode, default_ansi = False, True

default_formatting_config = {
    'unicode'             : default_unicode,
    'ansi'                : default_ansi,
    'emoji'               : {
        'hand'            : '👋',
        'error'           : '🛑',
        'failure'         : '💢',
        'success'         : '🎉',
        'warning'         : '⚠️',
        'info'            : '💬',
        'debug'           : '🐞',
        'question'        : '❓',
        'connector'       : '🔌',
        'metric'          : '📊',
        'location'        : '📍',
        'key'             : '🔑',
        'idea'            : '💡',
    },
    'warnings'            : {
        'unicode'         : {
            'icon'        : 'MRSM{formatting:emoji:warning}',
        },
        'ascii'           : {
            'icon'        : 'WARNING',
        },
        'ansi'            : {
            'rich'        : {
                'style'   : 'bold yellow',
            },
            #  'color'       : [
                #  'bold',
                #  'yellow',
            #  ],
        },
    },
    'success'             : {
        'unicode'         : {
            'icon'        : 'MRSM{formatting:emoji:success}',
        },
        'ascii'           : {
            'icon'        : '+',
        },
        'ansi'            : {
            'rich'        : {
                'style'   : 'bold bright_green',
            },
            #  'color'       : [
                #  'bold',
                #  'bright green',
            #  ],
        },
    },
    'failure'             : {
        'unicode'         : {
            'icon'        : 'MRSM{formatting:emoji:failure}',
        },
        'ascii'           : {
            'icon'        : '-',
        },
        'ansi'            : {
            'rich'        : {
                'style'   : 'bold red',
            },
            #  'color'       : [
                #  'bold',
                #  'red',
            #  ],
        },
    },
    'errors'              : {
        'unicode'         : {
            'icon'        : 'MRSM{formatting:emoji:error}',
        },
        'ascii'           : {
            'icon'        : 'ERROR',
        },
        'ansi'            : {
            'rich'        : {
                'style'   : 'bold red',
            },
            #  'color'       : [
                #  'bold',
                #  'red',
            #  ],
        },
    },
    'info'                : {
        'unicode'         : {
            'icon'        : 'MRSM{formatting:emoji:info}',
        },
        'ascii'           : {
            'icon'        : 'INFO',
        },
        'ansi'            : {
            'rich'        : {
                'style'   : 'bright_magenta',
            },
            #  'color'       : [
                #  'bright magenta',
            #  ],
        },
    },
    'question'            : {
        'unicode'         : {
            'icon'        : 'MRSM{formatting:emoji:question}',
        },
        'ascii'           : {
            'icon'        : '',
        },
        'ansi'            : {
            'rich'        : {
                'style'   : 'green',
            },
            #  'color'       : [
                #  'green',
            #  ],
        },
    },
    'debug'               : {
        'unicode'         : {
            'icon'        : 'MRSM{formatting:emoji:debug}',
        },
        'ascii'           : {
            'icon'        : 'DEBUG',
        },
        'ansi'            : {
            'rich'        : {
                'style'   : 'cyan',
            },
            #  'color'       : [
                #  'cyan',
            #  ],
        },
    },

}



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Show information about a Pipe
"""

from meerschaum.utils.typing import SuccessTuple

def show(
        self,
        nopretty : bool = False,
        debug : bool = False,
        **kw
    ) -> SuccessTuple:
    """
    Show attributes of a Pipe.
    """
    import json
    from meerschaum.utils.formatting import pprint, make_header
    from meerschaum.utils.warnings import info
    if not nopretty:
        print(make_header(f"Attributes for pipe '{self}':"))
        pprint(self.attributes)
    else:
        print(json.dumps(self.attributes))

    return True, "Success"



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Register new Pipes
"""
from meerschaum.utils.packages import attempt_import
pydantic = attempt_import('pydantic', warn=False)

class MetaPipe(pydantic.BaseModel):
    connector_keys : str ### e.g. sql:main
    metric_key : str
    location_key : str = None
    parameters : dict = None




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Layout for the error page.
"""

import dash_core_components as dcc
import dash_html_components as html
import dash_bootstrap_components as dbc
from meerschaum.api import endpoints

layout = dbc.Container([
    dcc.Markdown(f"""
    ## Error encountered
    You might have requested an invalid resource.
    Head back to the [dashboard home page]({endpoints['dash']}) and try navigating from there.
    """),
], className='jumbotron')



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Define JWT authorization here.
"""

import os
from meerschaum.api import app, endpoints
from meerschaum.utils.packages import attempt_import
fastapi = attempt_import('fastapi', lazy=False)
fastapi_responses = attempt_import('fastapi.responses', lazy=False)
fastapi_login = attempt_import('fastapi_login')

LoginManager = fastapi_login.LoginManager
def generate_secret_key() -> str:
    """
    Read or generate the secret keyfile.
    """
    from meerschaum.config._paths import API_SECRET_KEY_PATH
    if not API_SECRET_KEY_PATH.exists():
        secret_key = os.urandom(24).hex()
        with open(API_SECRET_KEY_PATH, 'w+') as f:
            f.write(secret_key)
    else:
        with open(API_SECRET_KEY_PATH, 'r') as f:
            secret_key = f.read()

    return secret_key

SECRET = generate_secret_key()
manager = LoginManager(SECRET, token_url=endpoints['login'])


#  fastapi_jwt_auth = attempt_import('fastapi_jwt_auth', lazy=False)
#  pydantic = attempt_import('pydantic', warn=False, lazy=False)
#  AuthJWT = fastapi_jwt_auth.AuthJWT
#  from fastapi_jwt_auth.exceptions import AuthJWTException


#  class JWTSettings(pydantic.BaseModel):
    #  """
    #  Settings for FastAPI JWT.
    #  """
    #  authjwt_secret_key: str = SECRET

#  @AuthJWT.load_config
#  def get_jwt_config():
    #  """
    #  Required for AuthJWT.
    #  """
    #  return JWTSettings()

#  @app.exception_handler(AuthJWTException)
#  def authjwt_exception_handler(request: fastapi.Request, exception: AuthJWTException):
    #  """
    #  Return JSON when JWT exception are encountered.
    #  """
    #  return fastapi_responses.JSONResponse(
        #  status_code = exception.status_code,
        #  content = { 'detail' : exception.message }
    #  )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Custom components are defined here.
"""

from __future__ import annotations
from dash.dependencies import Input, Output, State
from meerschaum.utils.packages import attempt_import
from meerschaum.utils.typing import SuccessTuple
from meerschaum.config.static import _static_config
from meerschaum.utils.misc import remove_ansi
from meerschaum.actions import get_shell
dbc = attempt_import('dash_bootstrap_components', lazy=False)
dcc = attempt_import('dash_core_components', warn=False)
dex = attempt_import('dash_extensions', lazy=False)
dash_ace = attempt_import('dash_ace', lazy=False)
html = attempt_import('dash_html_components', warn=False)

component_ids = {

}

keyboard = dex.Keyboard(
    id = 'keyboard',
    captureKeys = ['Enter'],
)
go_button = dbc.Button('Execute', id='go-button', color='primary', style={'width': '100%'})
test_button = dbc.Button('Test', id='test-button', color='danger', style={'display' : 'none'})
get_items_menu = dbc.DropdownMenu(
    label='Get Items', id='get-items-menu', children=[
        dbc.DropdownMenuItem("Pipes", id='get-pipes-button'),
        dbc.DropdownMenuItem("Jobs", id='get-jobs-button'),
    ],
    style={'width': '100%'}
)
cancel_button = dbc.Button('Cancel', id='cancel-button', color='danger', style={'width': '100%'})
bottom_buttons_content = dbc.Card(
    dbc.CardBody(
        dbc.Row([
            dbc.Col(go_button, width='2'),
            dbc.Col(cancel_button, width='2'),
            dbc.Col(get_items_menu, width='2'),
        ], no_gutters=False)
    )
)
console_div = html.Div(id='console-div', children=[html.Pre(get_shell().intro, id='console-pre')])

location = dcc.Location(id='location', refresh=False)
websocket = dex.WebSocket(id='ws', url="")

search_parameters_editor = dash_ace.DashAceEditor(
    id = 'search-parameters-editor',
    theme = 'monokai',
    mode = 'json',
    tabSize = 2,
    placeholder = (
        'Additional search parameters. ' +
        'Simple dictionary format or JSON accepted.'
    ),
    style = {'height' : 100},
)

def alert_from_success_tuple(success : SuccessTuple):
    """
    Return an Alert from a SuccessTuple.
    """
    return dbc.Alert('', is_open=False) if not isinstance(success, tuple) else (
        dbc.Alert(
            remove_ansi(success[1]),
            id = 'success-alert',
            dismissable = True,
            fade = True,
            is_open = not (success[1] in _static_config()['system']['success']['ignore']),
            color = 'success' if success[0] else 'danger',
        )
    )




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
### Load version and docs metadata


from meerschaum.config import __version__
from meerschaum._internal.docs import index as __doc__

#  from meerschaum.utils.packages import lazy_import
from meerschaum.Pipe import Pipe
from meerschaum.utils import get_pipes
from meerschaum.connectors import get_connector





#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Format ANSI text to HTML.
"""

from meerschaum.utils.packages import attempt_import
ansi2html = attempt_import('ansi2html')
ansi2html.converter._html_template = """


"""


def ansi_to_html(ansi : str) -> str:
    """
    Convert a string with ANSI codes to a stylized HTML span.
    """
    converter = ansi2html.Ansi2HTMLConverter()




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Functions for managing packages and virtual environments reside here.
"""

from __future__ import annotations
import importlib, os
from meerschaum.utils.typing import Any, List, SuccessTuple, Optional, Union

from meerschaum.utils.packages._packages import packages, all_packages
from meerschaum.utils.warnings import warn, error

_import_module = importlib.import_module
_import_hook_venv = None
active_venvs = set()
def need_update(
        package : 'ModuleType',
        split : bool = True,
        color : bool = True,
        debug : bool = False,
    ) -> bool:
    """
    Check if a Meerschaum dependency needs an update.
    """
    if debug:
        from meerschaum.utils.debug import dprint
    import re
    root_name = package.__name__.split('.')[0] if split else package.__version__
    install_name = all_packages.get(root_name, root_name)

    _install_no_version = re.split('[=<>,! ]', install_name)[0]
    if debug:
        dprint(f"_install_no_version: {_install_no_version}", color=color)
    required_version = install_name.replace(_install_no_version, '')
    if debug:
        dprint(f"required_version: {required_version}", color=color)
    update_checker = attempt_import('update_checker', lazy=False, check_update=False)
    checker = update_checker.UpdateChecker()
    try:
        version = package.__version__
    except Exception as e:
        return False

    result = checker.check(_install_no_version, version)
    if result is None:
        return False
    if required_version:
        semver = attempt_import('semver')
        if debug:
            dprint(f"Available version: {result.available_version}", color=color)
            dprint(f"Required version: {required_version}", color=color)
        try:
            return semver.match(required_version, result.available_version)
        except Exception as e:
            return False

    packaging_version = attempt_import('packaging.version')
    return (
        packaging_version.parse(result.available_version) > 
        packaging_version.parse(version)
    )

def is_venv_active(
        venv: str = 'mrsm',
        color : bool = True,
        debug: bool = False
    ) -> bool:
    """
    Check if a virtual environment is active
    """
    if venv is None:
        return False
    if debug:
        from meerschaum.utils.debug import dprint
        dprint(f"Checking if virtual environment '{venv}' is active.", color=color)
    return venv in active_venvs

def deactivate_venv(
        venv: str = 'mrsm',
        color : bool = True,
        debug: bool = False
    ) -> bool:
    """
    Remove a virtual environment from sys.path (if it's been activated)
    """
    import sys
    global active_venvs

    if venv is None:
        return True

    if debug:
        from meerschaum.utils.debug import dprint
        dprint(f"Deactivating virtual environment '{venv}'...", color=color)

    if venv in active_venvs:
        active_venvs.remove(venv)

    if sys.path is None:
        return False

    target = venv_target_path(venv, debug=debug)
    if str(target) in sys.path:
        sys.path.remove(str(target))

    if debug:
        dprint(f'sys.path: {sys.path}', color=color)

    return True

def activate_venv(
        venv: str = 'mrsm',
        color : bool = True,
        debug: bool = False
    ) -> bool:
    """
    Create a virtual environment (if it doesn't exist) and add it to sys.path if necessary
    """
    global active_venvs
    if venv in active_venvs or venv is None:
        return True
    if debug:
        from meerschaum.utils.debug import dprint
    import sys, os, platform, pathlib
    from meerschaum.config._paths import VIRTENV_RESOURCES_PATH
    _venv = None
    virtualenv = attempt_import(
        'virtualenv', venv=None, lazy=False, install=True, warn=False, debug=debug, color=False,
    )
    if virtualenv is None:
        try:
            import ensurepip
            import venv as _venv
            virtualenv = None
        except ImportError:
            _venv = None
    if virtualenv is None and _venv is None:
        print(
            "Failed to import virtualenv! "
            + "Please install virtualenv via pip then restart Meerschaum."
        )
        sys.exit(1)
    venv_path = VIRTENV_RESOURCES_PATH / venv
    bin_path = pathlib.Path(
        venv_path,
        ('bin' if platform.system() != 'Windows' else "Scripts")
    )
    if not venv_path.exists():
        if _venv is not None:
            _venv.create(
                venv_path,
                system_site_packages = False,
                with_pip = True,
                symlinks = (platform.system() != 'Windows'),
            )
        else:
            virtualenv.cli_run([str(venv_path), '--download', '--system-site-packages'])

    old_cwd = pathlib.Path(os.getcwd())
    os.chdir(VIRTENV_RESOURCES_PATH)
    try:
        if debug:
            dprint(f"Activating virtual environment '{venv}'...", color=color)
        active_venvs.add(venv)
    except Exception:
        pass
    finally:
        os.chdir(old_cwd)

    ### override built-in import with attempt_import
    #  install_import_hook(venv, debug=debug)
    target = venv_target_path(venv, debug=debug)
    if str(target) not in sys.path:
        sys.path.insert(0, str(target))
    if debug:
        dprint(f'sys.path: {sys.path}', color=color)
    #  print(sys.path)
    return True

def venv_exec(code: str, venv: str = 'mrsm', debug: bool = False) -> bool:
    """
    Execute Python code in a subprocess via a virtual environment's interpeter.

    Return True if the code successfully executes, False on failure.
    """
    import subprocess, sys, platform, os
    from meerschaum.config._paths import VIRTENV_RESOURCES_PATH
    from meerschaum.utils.debug import dprint
    executable = (
        sys.executable if venv is None
        else os.path.join(
            VIRTENV_RESOURCES_PATH, venv, (
                'bin' if platform.system() != 'Windows' else 'Scripts'
            ), 'python'
        )
    )
    cmd_list = [executable, '-c', code]
    if debug:
        dprint(str(cmd_list))
    return subprocess.call(cmd_list) == 0

def get_pip(debug : bool = False) -> bool:
    """
    Download and run the get-pip.py script.
    """
    import sys, subprocess
    from meerschaum.utils.misc import wget
    from meerschaum.config._paths import CACHE_RESOURCES_PATH
    from meerschaum.config.static import _static_config
    url = _static_config()['system']['urls']['get-pip.py']
    dest = CACHE_RESOURCES_PATH / 'get-pip.py'
    try:
        wget(url, dest, color=False, debug=debug)
    except Exception as e:
        print(f"Failed to fetch pip from '{url}'. Please install pip and restart Meerschaum.") 
        sys.exit(1)
    cmd_list = [sys.executable, str(dest)] 
    return subprocess.call(cmd_list) == 0

def pip_install(
        *packages: List[str],
        args: Optional[List[str]] = None,
        venv: str = 'mrsm',
        deactivate: bool = True,
        split : bool = False,
        check_update : bool = True,
        check_wheel : bool = True,
        _uninstall : bool = False,
        color : bool = True,
        debug: bool = False
    ) -> bool:
    """
    Install pip packages
    """
    from meerschaum.config.static import _static_config
    if args is None:
        args = ['--upgrade'] if not _uninstall else []
    if color:
        try:
            from meerschaum.utils.formatting import ANSI, UNICODE
        except ImportError:
            ANSI, UNICODE = False, False
    else:
        ANSI, UNICODE = False, False
    if check_wheel:
        try:
            import wheel
            have_wheel = True
        except ImportError as e:
            have_wheel = False
    _args = list(args)
    try:
        if venv is not None:
            activate_venv(venv=venv, color=color, debug=debug)
        import pip
        have_pip = True
        if venv is not None and deactivate:
            deactivate_venv(venv=venv, debug=debug, color=color)
    except ImportError:
        have_pip = False
    if not have_pip:
        try:
            import ensurepip
        except ImportError:
            ensurepip = None
        if ensurepip is None:
            if not get_pip(debug=debug):
                import sys
                print(
                    "Failed to import pip and ensurepip. " +
                    "Please install pip and restart Meerschaum.\n\n" +
                    "You can find instructions on installing pip here: " +
                    "https://pip.pypa.io/en/stable/installing/"
                )
                sys.exit(1)
        else:
            ensurepip.bootstrap(upgrade=True, )
        import pip
    if venv is not None:
        activate_venv(venv=venv, debug=debug, color=color)
        if '--ignore-installed' not in args and '-I' not in _args and not _uninstall:
            _args += ['--ignore-installed']

    if check_update and need_update(pip, debug=debug) and not _uninstall:
        _args.append('pip')
    _args = (['install'] if not _uninstall else ['uninstall']) + _args

    if check_wheel and not _uninstall:
        if not have_wheel:
            _args.append('wheel')

    if not ANSI and '--no-color' not in _args:
        _args.append('--no-color')

    if '--no-warn-conflicts' not in _args and not _uninstall:
        _args.append('--no-warn-conflicts')

    if '--disable-pip-version-check' not in _args:
        _args.append('--disable-pip-version-check')

    if venv is not None and '--target' not in _args and '-t' not in _args and not _uninstall:
        _args += ['--target', venv_target_path(venv, debug=debug)]
    elif (
        '--target' not in _args
            and '-t' not in _args
            and os.environ.get(
                _static_config()['environment']['runtime'], None
            ) != 'docker'
            and not inside_venv()
            and not _uninstall
    ):
        _args += ['--user']
    if '--progress-bar' in _args:
        _args.remove('--progress-bar')
    if UNICODE and not _uninstall:
        _args += ['--progress-bar', 'pretty']
    elif not _uninstall:
        _args += ['--progress-bar', 'ascii']
    if debug:
        if '-v' not in _args or '-vv' not in _args or '-vvv' not in _args:
            pass
    else:
        if '-q' not in _args or '-qq' not in _args or '-qqq' not in _args:
            pass

    _packages = []
    for p in packages:
        root_name = p.split('.')[0] if split else p
        install_name = all_packages.get(root_name, root_name)
        _packages.append(install_name)

    msg = "Installing packages:" if not _uninstall else "Uninstalling packages:"
    for p in _packages:
        msg += f'\n  - {p}'
    print(msg)

    success = run_python_package('pip', _args + _packages, venv=venv, debug=debug) == 0
    if venv is not None and deactivate:
        deactivate_venv(venv=venv, debug=debug, color=color)
    msg = (
        "Successfully " + ('un' if _uninstall else '') + "installed packages." if success 
        else "Failed to " + ('un' if _uninstall else '') + "install packages."
    )
    print(msg)
    if debug:
        print('pip ' + ('un' if _uninstall else '') + 'install returned:', success)
    return success

def pip_uninstall(
        *args, **kw
    ) -> bool:
    """
    Uninstall Python packages.
    """
    return pip_install(*args, _uninstall=True, **{k: v for k, v in kw.items() if k != '_uninstall'})

def run_python_package(
        package_name : str,
        args : Optional[List[str]] = None,
        venv : Optional[str] = None,
        cwd : Optional[str] = None,
        foreground : bool = False,
        debug : bool = False,
        **kw : Any,
    ) -> Union[int, subprocess.Popen]:
    """
    Runs an installed python package.
    E.g. Translates to `/usr/bin/python -m [package]`

    :param package_name:
        The Python module to be executed.

    :param args:
        Additional command line arguments to be appended after `-m [package]`.
        Defaults to `None`.

    :param venv:
        If specified, execute the Python interpreter from a virtual environment.
        Defaults to `None`.

    :param cwd:
        If specified, change directories before starting the process.
        Defaults to `None`.

    :param foreground:
        If `True`, start the subprocess as a foreground process.
        Defaults to `False`.

    :param kw:
        Additional keyword arguments to pass to `meerschaum.utils.process.run_process()`
        and by extension `subprocess.Popen()`.
    """
    import sys, platform
    from subprocess import call
    from meerschaum.config._paths import VIRTENV_RESOURCES_PATH
    from meerschaum.utils.process import run_process
    from meerschaum.utils.warnings import warn
    if args is None:
        args = []
    old_cwd = os.getcwd()
    if cwd is not None:
        os.chdir(cwd)
    executable = (
        sys.executable if venv is None
        else os.path.join(
            VIRTENV_RESOURCES_PATH, venv, (
                'bin' if platform.system() != 'Windows' else 'Scripts'
            ), ('python' + ('.exe' if platform.system() == 'Windows' else ''))
        )
    )
    command = [executable, '-m', str(package_name)] + [str(a) for a in args]
    import traceback
    if debug:
        print(command, file=sys.stderr)
    try:
        rc = run_process(command, foreground=foreground, **kw)
    except Exception as e:
        print(traceback.format_exc(e))
        warn(e, color=False)
        rc = call(command)
    except KeyboardInterrupt:
        rc = 1
    os.chdir(old_cwd)
    return rc

def attempt_import(
        *names: List[str],
        lazy: bool = True,
        warn: bool = True,
        install: bool = True,
        venv: str = 'mrsm',
        precheck: bool = True,
        split: bool = True,
        check_update : bool = False,
        deactivate : bool = True,
        color : bool = True,
        debug: bool = False
    ) -> Union['ModuleType', Tuple['ModuleType']]:
    """
    Raise a warning if packages are not installed; otherwise import and return modules.
    If lazy = True, return lazy-imported modules.

    Returns tuple of modules if multiple names are provided, else returns one module.

    Examples:

        ```
        >>> pandas, sqlalchemy = attempt_import('pandas', 'sqlalchemy')
        >>> pandas = attempt_import('pandas')

        ```

    :param names:
        The packages to be imported.

    :param lazy:
        If True, lazily load packages.
        Defaults to False.

    :param warn:
        If True, raise a warning if a package cannot be imported.
        Defaults to True.

    :param install:
        If True, attempt to install a missing package into the designated virtual environment.
        If `check_update` is True, install updates if available.
        Defaults to True.

    :param venv:
        The virtual environment in which to search for packages and to install packages into.
        Defaults to 'mrsm'.

    :param precheck:
        If True, attempt to find module before importing (necessary for checking if modules exist
        and retaining lazy imports), otherwise assume lazy is False.
        Defaults to True.        

    :param split:
        If True, split packages' names on '.'.
        Defaults to True.

    :param check_update:
        If True, check PyPI for the most recent version.
        If `install` is True, install updates if available.
        Defaults to False.

    """

    import importlib.util

    ### to prevent recursion, check if parent Meerschaum package is being imported
    if names == ('meerschaum',):
        return _import_module('meerschaum')

    if venv == 'mrsm' and _import_hook_venv is not None:
        if debug:
            print(f"Import hook for virtual environment '{_import_hook_venv}' is active.")
        venv = _import_hook_venv

    if venv is not None:
        activate_venv(venv=venv, color=color, debug=debug)
    _warnings = _import_module('meerschaum.utils.warnings')
    warn_function = _warnings.warn

    def do_import(_name: str, **kw):
        if venv is not None:
            activate_venv(venv=venv, debug=debug)
        ### determine the import method (lazy vs normal)
        from meerschaum.utils.misc import filter_keywords
        import_method = _import_module if not lazy else lazy_import
        try:
            mod = import_method(_name, **(filter_keywords(import_method, **kw)))
        except Exception as e:
            if warn:
                warn_function(
                    f"Failed to import module '{_name}'.\nException:\n{e}",
                    ImportWarning,
                    stacklevel = (5 if lazy else 4),
                    color = False,
                )
            mod = None
        if venv is not None:
            deactivate_venv(venv=venv, color=color, debug=debug)
        return mod

    modules = []
    for name in names:
        ### Enforce virtual environment (something is deactivating in the loop so check each pass).
        if venv is not None:
            activate_venv(debug=debug)
        ### Check if package is a declared dependency.
        root_name = name.split('.')[0] if split else name
        install_name = all_packages.get(root_name, None)
        if install_name is None:
            install_name = root_name
            if warn:
                warn_function(
                    f"Package '{root_name}' is not declared in meerschaum.utils.packages.",
                    ImportWarning,
                    stacklevel = 3,
                    color = False
                )

        ### Determine if the package exists.
        if precheck is False:
            found_module = (
                do_import(
                    name, debug=debug, warn=False, venv=venv, color=color
                ) is not None
            )
        else:
            try:
                found_module = (importlib.util.find_spec(name) is not None)
            except ModuleNotFoundError as e:
                found_module = False

        if not found_module:
            if install:
                ### NOTE: pip_install deactivates venv, so deactivate must be False.
                if not pip_install(
                    root_name,
                    venv = venv,
                    deactivate = False,
                    split = False,
                    #  split = split,
                    check_update = check_update,
                    color = color,
                    debug = debug
                ) and warn:
                    warn_function(
                        f"Failed to install '{install_name}'.",
                        ImportWarning,
                        stacklevel = 3,
                        color = False, ### Color triggers a deactivate, so keep as False
                    )
            elif warn:
                ### Raise a warning if we can't find the package and install = False.
                warn_function(
                    (f"\n\nMissing package '{name}'; features will not work correctly. "
                     f"\n\nSet install=True when calling attempt_import.\n"),
                    ImportWarning,
                    stacklevel = 3,
                    color = False,
                )

        ### Do the import. Will be lazy if lazy=True.
        m = do_import(name, debug=debug, warn=warn, venv=venv, color=color)
        modules.append(m)

        ### Check for updates (skip this by default)
        if check_update:
            if need_update(m, split=split, debug=debug):
                if install:
                    if not pip_install(
                        root_name,
                        venv = venv,
                        split = False,
                        deactivate = False,
                        check_update = check_update,
                        color = color,
                        debug = debug
                    ) and warn:
                        warn_function(
                            f"There's an update available for '{install_name}', " +
                            "but it failed to install. " +
                            "Try install via Meershaum with `install packages '{install_name}'`.",
                            ImportWarning,
                            stacklevel = 3,
                            color = False,
                        )
                elif warn:
                    warn_function(
                        f"There's an update available for '{m.__name__}'.",
                        stack = False,
                        color = False,
                    )
    if venv is not None and deactivate:
        deactivate_venv(venv=venv, debug=debug, color=color)

    modules = tuple(modules)
    if len(modules) == 1:
        return modules[0]
    return modules

def lazy_import(
        name: str,
        local_name: str = None,
        venv : Optional[str] = None,
        warn : bool = True, 
        deactivate : bool = True,
        debug : bool = False
    ) -> meerschaum.utils.packages.lazy_loader.LazyLoader:
    """
    Lazily import a package
    Uses the tensorflow LazyLoader implementation (Apache 2.0 License)
    """
    from meerschaum.utils.packages.lazy_loader import LazyLoader
    if local_name is None:
        local_name = name
    return LazyLoader(
        local_name,
        globals(),
        name,
        venv = venv,
        warn = warn,
        deactivate = deactivate,
        debug = debug
    )

def import_pandas(deactivate : bool = False, debug : bool = False, lazy : bool = False,  **kw) -> 'ModuleType':
    """
    Quality-of-life function to attempt to import the configured version of pandas.
    """
    import sys
    from meerschaum.config import get_config
    pandas_module_name = get_config('system', 'connectors', 'all', 'pandas', patch=True)
    ### NOTE: modin does NOT currently work!
    if pandas_module_name == 'modin':
        pandas_module_name = 'modin.pandas'

    activate_venv(venv='mrsm', debug=debug)
    pd = attempt_import(pandas_module_name, deactivate=deactivate, debug=debug, lazy=lazy, **kw)
    if deactivate:
        deactivate_venv(venv='mrsm', debug=debug)
    return pd

def import_rich(
        lazy: bool = True, deactivate : bool = False, debug : bool = False,
        **kw : Any
    ) -> 'ModuleType':
    """
    Quality of life function for importing rich.
    """
    from meerschaum.utils.formatting import ANSI, UNICODE
    if not ANSI and not UNICODE:
        return None

    activate_venv(venv='mrsm', debug=debug)

    ## need typing_extensions for `from rich import box`
    typing_extensions = attempt_import('typing_extensions', deactivate=deactivate, lazy=False)
    pygments = attempt_import('pygments', deactivate=deactivate, lazy=False)
    rich = attempt_import('rich', lazy=lazy, deactivate=deactivate, **kw)
    if deactivate:
        deactivate_venv(venv='mrsm', debug=debug)
    return rich

def get_modules_from_package(
        package: 'package',
        names: bool = False,
        recursive: bool = False,
        lazy: bool = False,
        modules_venvs: bool = False,
        debug: bool = False
    ):
    """
    Find and import all modules in a package.

    Returns: either list of modules or tuple of lists

    names = False (default) : modules
    names = True            : (__all__, modules)
    """
    from os.path import dirname, join, isfile, isdir, basename
    import glob

    pattern = '*' if recursive else '*.py'
    module_names = glob.glob(join(dirname(package.__file__), pattern), recursive=recursive)
    _all = [
        basename(f)[:-3] if isfile(f) else basename(f)
        for f in module_names
            if ((isfile(f) and f.endswith('.py')) or isdir(f))
               and not f.endswith('__init__.py')
               and not f.endswith('__pycache__')
    ]

    if debug:
        from meerschaum.utils.debug import dprint
        dprint(str(_all))
    modules = []
    for module_name in [package.__name__ + "." + mod_name for mod_name in _all]:
        ### there's probably a better way than a try: catch but it'll do for now
        try:
            ### if specified, activate the module's virtual environment before importing.
            ### NOTE: this only considers the filename, so two modules from different packages
            ### may end up sharing virtual environments.
            if modules_venvs:
                activate_venv(module_name.split('.')[-1], debug=debug)
            m = lazy_import(module_name, debug=debug) if lazy else _import_module(module_name)
            modules.append(m)
        except Exception as e:
            if debug:
                dprint(e)
        finally:
            if modules_venvs:
                deactivate_venv(module_name.split('.')[-1], debug=debug)
    if names:
        return _all, modules

    return modules

def import_children(
        package: Optional['ModuleType'] = None,
        package_name: Optional[str] = None,
        types : Optional[List[str]] = None,
        lazy: bool = True,
        recursive: bool = False,
        debug: bool = False
    ) -> List['ModuleType']:
    """
    Import all functions in a package to its __init__.
    Returns of list of modules.

    :param package:
        Package to import its functions into.
        If None (default), use parent.

    :param package_name:
        Name of package to import its functions into
        If None (default), use parent.

    :param types:
        Types of members to return.
        Default : ['method', 'builtin', 'class', 'function', 'package', 'module']

    """
    import sys, inspect

    if types is None:
        types = ['method', 'builtin', 'function', 'class', 'module']

    ### if package_name and package are None, use parent
    if package is None and package_name is None:
        package_name = inspect.stack()[1][0].f_globals['__name__']

    ### populate package or package_name from other other
    if package is None:
        package = sys.modules[package_name]
    elif package_name is None:
        package_name = package.__name__

    ### Set attributes in sys module version of package.
    ### Kinda like setting a dictionary
    ###   functions[name] = func
    modules = get_modules_from_package(package, recursive=recursive, lazy=lazy, debug=debug)
    _all, members = [], []
    objects = []
    for module in modules:
        _objects = []
        for ob in inspect.getmembers(module):
            for t in types:
                ### ob is a tuple of (name, object)
                if getattr(inspect, 'is' + t)(ob[1]):
                    _objects.append(ob)

        if 'module' in types:
            _objects.append((module.__name__.split('.')[0], module))
        objects += _objects
    for ob in objects:
        setattr(sys.modules[package_name], ob[0], ob[1])
        _all.append(ob[0])
        members.append(ob[1])

    if debug:
        from meerschaum.utils.debug import dprint
        dprint(str(_all))
    ### set __all__ for import *
    setattr(sys.modules[package_name], '__all__', _all)
    return members

def reload_package(
        package: str,
        lazy: bool = False,
        debug: bool = False,
        **kw: Any
    ) -> 'ModuleType':
    """
    Recursively load a package's subpackages, even if they were not previously loaded
    """
    import pydoc
    if isinstance(package, str):
        package_name = package
    else:
        try:
            package_name = package.__name__
        except Exception as e:
            package_name = str(package)
    return pydoc.safeimport(package_name, forceload=1)

def is_installed(
        name: str,
        venv : Optional[str] = None,
    ) -> bool:
    """
    Check whether a package is installed.
    name : str
        Name of the package in question
    """
    import importlib.util
    if venv is not None:
        activate_venv(venv=venv)
    found = importlib.util.find_spec(name) is not None
    if venv is None:
        return found
    mod = attempt_import(name, venv=venv)
    found = (venv == package_venv(mod))
    deactivate_venv(venv=venv)
    return found

def venv_contains_package(
        name : str, venv : Optional[str] = None
    ) -> bool:
    """
    Search the contents of a virtual environment for a package.
    """
    import os
    if venv is None:
        return is_installed(name, venv=venv)
    vtp = venv_target_path(venv)
    if not vtp.exists():
        return False
    return name in os.listdir(venv_target_path(venv))

def venv_target_path(venv : str, debug : bool = False) -> pathlib.Path:
    """
    Return a virtual environment's site-package path.
    """
    import os, sys, platform, pathlib
    from meerschaum.config._paths import VIRTENV_RESOURCES_PATH

    venv_root_path = str(os.path.join(VIRTENV_RESOURCES_PATH, venv))
    target_path = venv_root_path

    ### Ensure 'lib' or 'Lib' exists.
    lib = 'lib' if platform.system() != 'Windows' else 'Lib'
    if lib not in os.listdir(venv_root_path):
        print(f"Failed to find lib directory for virtual environment '{venv}'.")
        sys.exit(1)
    target_path = os.path.join(target_path, lib)

    ### Check if a 'python3.x' folder exists.
    python_folder = 'python' + str(sys.version_info.major) + '.' + str(sys.version_info.minor)
    if python_folder in os.listdir(target_path): ### Linux
        target_path = os.path.join(target_path, python_folder)

    ### Ensure 'site-packages' exists.
    if 'site-packages' in os.listdir(target_path): ### Windows
        target_path = os.path.join(target_path, 'site-packages')
    else:
        from meerschaum.config._paths import set_root
        import traceback
        traceback.print_stack()
        print(f"Failed to find site-packages directory for virtual environment '{venv}'.")
        sys.exit(1)

    if debug:
        print(f"Target path for virtual environment '{venv}':\n" + str(target_path))
    return pathlib.Path(target_path)

def package_venv(package : 'ModuleType') -> Optional[str]:
    """
    Inspect a package and return the virtual environment in which it presides.
    """
    from meerschaum.config._paths import VIRTENV_RESOURCES_PATH
    if str(VIRTENV_RESOURCES_PATH) not in package.__file__:
        return None
    return package.__file__.split(str(VIRTENV_RESOURCES_PATH))[1].split(os.path.sep)[1]

def inside_venv() -> bool:
    """
    Determine whether current Python interpreter is running inside a virtual environment.
    """
    import sys
    return (
        hasattr(sys, 'real_prefix') or (
            hasattr(sys, 'base_prefix')
                and sys.base_prefix != sys.prefix
        )
    )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Create connectors with `meerschaum.connectors.get_connector()`.
For ease of use, you can also import from the root `meerschaum` module:
```
>>> from meerschaum import get_connector
>>> conn = get_connector()
```
"""

from __future__ import annotations
from meerschaum.utils.typing import Any, SuccessTuple, Union, Optional, Sequence, Mapping

from meerschaum.connectors.Connector import Connector
from meerschaum.connectors.sql._create_engine import flavor_configs as sql_flavor_configs

### store connectors partitioned by
### type, label for reuse
connectors = {
    'api'    : dict(),
    'sql'    : dict(),
    'mqtt'   : dict(),
    'plugin' : dict(),
}
attributes = {
    'api' : {
        'required' : [
            'host',
            'username',
            'password'
        ],
        'default' : {
            #  'username' : 'mrsm',
            #  'password' : 'mrsm',
            'protocol' : 'http',
            'port'     : 8000,
        },
    },
    'sql' : {
        'flavors' : sql_flavor_configs,
    },
    'mqtt' : {
        'required' : [
            'host',
        ],
        'default' : {
            'port'     : 1883,
            'keepalive': 60,
        },
    },
}
### fill this with classes only on execution
### for lazy loading
types = dict()

def get_connector(
        type : str = None,
        label : str = None,
        refresh : bool = False,
        debug : bool = False,
        **kw : Any
    ):
    """
    Return existing connector or create new connection and store for reuse.

    You can create new connectors if enough parameters are provided for the given type and flavor.

    For example, the following parameters would create a new SQLConnector that isn't in the configuration file.
    
    ```
    >>> conn = get_connector(
    ...     type = 'sql',
    ...     label = 'newlabel',
    ...     flavor = 'sqlite',
    ...     database = '/file/path/to/database.db'
    ... )
    >>> 
    ```

    :param type:
        Connector type (sql, api, etc.). Defaults to the type of the configured `instance_connector`.
    
    :param label:
        Connector label (e.g. main). Defaults to `'main'`.

    :param refresh:
        Refresh the Connector instance / construct new object. Defaults to `False`.

    :param kw:
        Other arguments to pass to the Connector constructor.
        If the Connector has already been constructed and new arguments are provided,
        refresh is set to True and the old Connector is replaced.

    """
    from meerschaum.connectors.parse import parse_instance_keys
    from meerschaum.config import get_config
    from meerschaum.config.static import _static_config
    if type is None and label is None:
        default_instance_keys = get_config('meerschaum', 'instance', patch=True)
        ### recursive call to get_connector
        return parse_instance_keys(default_instance_keys)

    ### NOTE: the default instance connector may not be main.
    ### Only fall back to 'main' if the type is provided by the label is omitted.
    label = label if label is not None else _static_config()['connectors']['default_label']

    global types, connectors

    ### type might actually be a label. Check if so and raise a warning.
    if type not in connectors:
        possibilities, poss_msg = [], ""
        for _type in get_config('meerschaum', 'connectors'):
            if type in get_config('meerschaum', 'connectors', _type):
                possibilities.append(f"{_type}:{type}")
        if len(possibilities) > 0:
            poss_msg = " Did you mean"
            for poss in possibilities[:-1]:
                poss_msg += f" '{poss}',"
            if poss_msg.endswith(','):
                poss_msg = poss_msg[:-1]
            if len(possibilities) > 1:
                poss_msg += " or"
            poss_msg += f" '{possibilities[-1]}'?"

        from meerschaum.utils.warnings import warn
        warn(f"Cannot create Connector of type '{type}'." + poss_msg, stack=False)
        return None

    if len(types) == 0:
        from meerschaum.connectors.sql import SQLConnector
        from meerschaum.connectors.api import APIConnector
        from meerschaum.connectors.mqtt import MQTTConnector
        from meerschaum.connectors.plugin import PluginConnector
        types = {
            'api'    : APIConnector,
            'sql'    : SQLConnector,
            'mqtt'   : MQTTConnector,
            'plugin' : PluginConnector,
        }
    
    ### always refresh MQTT Connectors NOTE: test this!
    if type == 'mqtt':
        refresh = True

    ### determine if we need to call the constructor
    if not refresh:
        ### see if any user-supplied arguments differ from the existing instance
        if label in connectors[type]:
            warning_message = None
            for attribute, value in kw.items():
                if attribute not in connectors[type][label].__dict__:
                    warning_message = (
                        f"Received new attribute '{attribute}' not present in connector " +
                        f"{connectors[type][label]}.\n"
                    )
                elif connectors[type][label].__dict__[attribute] != value:
                    warning_message = (
                        f"Mismatched values for attribute '{attribute}' in connector '{connectors[type][label]}'.\n" +
                        f"  - Keyword value: '{value}'\n" +
                        f"  - Existing value: '{connectors[type][label].__dict__[attribute]}'\n"
                    )
            if warning_message is not None:
                from meerschaum.utils.warnings import warn
                warning_message += f"\nSetting `refresh` to True and recreating connector with type: '{type}' and label '{label}'"
                refresh = True
                warn(warning_message)
        else: ### connector doesn't yet exist
            refresh = True

    ### only create an object if refresh is True (can be manually specified, otherwise determined above)
    from meerschaum.utils.warnings import error
    import traceback
    error_msg = None
    if refresh:
        ### will raise an error if configuration is incorrect / missing
        conn = types[type](label=label, debug=debug, **kw)
        connectors[type][label] = conn

    return connectors[type][label]



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Modules, classes, and variables declared in _internal are not intended for general public use.
"""

#  from meerschaum.utils.packages import lazy_import

#  from meerschaum._internal.Plugin import Plugin
#  from meerschaum._internal.User import User



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Return a dictionary (or list) of pipe objects. See documentation below for more information.
"""

from __future__ import annotations

from meerschaum.utils.typing import (
    Sequence, Optional, Union, Mapping, Any, InstanceConnector, PipesDict, List, Dict, Tuple
)

def get_pipes(
        connector_keys : Optional[Union[str, Sequence[str]]] = None,
        metric_keys : Optional[Union[str, Sequence[str]]] = None,
        location_keys : Optional[Union[str, Sequence[str]]] = None,
        params : Mapping[str, Any] = dict(),
        mrsm_instance : Union[str, InstanceConnector, None] = None,
        as_list : bool = False,
        method : str = 'registered',
        wait : bool = False,
        debug : bool = False,
        **kw : Any
    ) -> Union[PipesDict, List['meerschaum.Pipe']]:
    """
    Return a dictionary (or list) of pipe objects.
    The pipes dictionary returned from `meerschaum.utils.get_pipes` has the following format:

    ```
    >>> pipes = {
    ...     <connector_keys> : {
    ...         <metric> : {
    ...             <location> : Pipe(
    ...                 <connector_keys>,
    ...                 <metric>,
    ...                 <location>,
    ...             ),
    ...         },
    ...     },
    ... },
    >>> 
    ```

    **NOTE:** If a pipe does not have a `location` key, it must be referenced as `None`.
    E.g. The pipe `sql:main_weather` would look like this:

    ```
    >>> pipes['sql:main']['weather'][None]
    ```

    :param connector_keys:
        String or list of connector keys.
        If parameter is omitted or is '*', fetch all connector_keys.

    :param metric_keys:
        String or list of metric keys.
        See connector_keys for formatting

    :param location_keys:
        String or list of location keys.
        See connector_keys for formatting

    :param params:
        Dictionary of additional parameters to search by.
        Params are parsed into a SQL WHERE clause.
        E.g. { 'a' : 1, 'b' : 2 } equates to 'WHERE a = 1 AND b = 2'

    :param mrsm_instance:
        Connector keys for the Meerschaum instance of the pipes.
        Must be of SQLConnector or APIConnector and point to a valid
        Meerschaum instance.

    :param as_list:
        If True, return pipes in a list instead of a hierarchical dictionary.
        False : { connector_keys : { metric_key : { location_key : Pipe } } }
        True  : [ Pipe ]
        Defaults to False.

    :param method:
        ['registered', 'explicit', 'all']

    If 'registered' (default), create pipes based on registered keys in the connector's pipes table
    (API or SQL connector, depends on mrsm_instance).
    
    If 'explicit', create pipes from provided connector_keys, metric_keys, and location_keys
    instead of consulting the pipes table. Useful for creating non-existent pipes.
    
    If 'all', create pipes from predefined metrics and locations. Required connector_keys.
    **NOTE:** Not implemented!

    :param wait:
        Wait for a connection before getting Pipes. Should only be true for cases where the
        database might not be running (like the API).
    """

    from meerschaum.config import get_config
    from meerschaum.utils.warnings import error

    if connector_keys is None:
        connector_keys = []
    if metric_keys is None:
        metric_keys = []
    if location_keys is None:
        location_keys = []

    if isinstance(connector_keys, str):
        connector_keys = [connector_keys]
    if isinstance(metric_keys, str):
        metric_keys = [metric_keys]
    if isinstance(location_keys, str):
        location_keys = [location_keys]

    ### Get SQL or API connector (keys come from `connector.fetch_pipes_keys()`).
    ### If `wait`, wait until a connection is made
    if mrsm_instance is None:
        mrsm_instance = get_config('meerschaum', 'instance', patch=True)
    if isinstance(mrsm_instance, str):
        from meerschaum.connectors.parse import parse_instance_keys
        connector = parse_instance_keys(keys=mrsm_instance, wait=wait, debug=debug)
    else: ### NOTE: mrsm_instance MUST be a SQL or API connector for this to work
        from meerschaum.connectors import Connector
        valid_connector = False
        if issubclass(type(mrsm_instance), Connector):
            if mrsm_instance.type in ('api', 'sql'):
                valid_connector = True
        if not valid_connector:
            error(f"Invalid instance connector: {mrsm_instance}")
        connector = mrsm_instance
    if debug:
        from meerschaum.utils.debug import dprint
        dprint(f"Using instance connector: {connector}")
    if not connector:
        error(f"Could not create connector from keys: '{mrsm_instance}'")

    ### get a list of tuples of keys based on the method type
    result = methods(
        method,
        connector,
        connector_keys = connector_keys,
        metric_keys = metric_keys,
        location_keys = location_keys,
        params = params,
        debug = debug
    )
    if result is None:
        error(f"Unable to build pipes!")

    ### populate the `pipes` dictionary with Pipes based on the keys
    ### obtained from the chosen `method`.
    from meerschaum import Pipe
    pipes = dict()
    for ck, mk, lk in result:
        if ck not in pipes:
            pipes[ck] = dict()

        if mk not in pipes[ck]:
            pipes[ck][mk] = dict()

        pipes[ck][mk][lk] = Pipe(ck, mk, lk, mrsm_instance=connector, debug=debug)

    if not as_list:
        return pipes
    from meerschaum.utils.misc import flatten_pipes_dict
    return flatten_pipes_dict(pipes)

def methods(
        method : str,
        connector : 'meerschaum.connectors.Connector',
        **kw : Any
    ) -> List[Tuple[str, str, str]]:
    """
    Return a list of tuples containing (connector_keys, metric_key, location_key)
    based on the method of choosing keys.
    """
    from meerschaum.utils.warnings import error
    from meerschaum.connectors.sql.tables import get_tables
    tables = get_tables(connector)

    def _registered(
            connector_keys : Optional[List[str]] = None,
            metric_keys : Optional[List[str]] = None,
            location_keys : Optional[List[str]] = None,
            params : Optional[Dict[str, Any]] = None,
            debug : bool = False,
            **kw
        ) -> List[Tuple[str, str, str]]:
        """
        Get keys from the pipes table or the API directly.
        Builds query or URL based on provided keys and parameters.

        Only works for SQL and API Connectors.
        """
        if connector_keys is None:
            connector_keys = []
        if metric_keys is None:
            metric_keys = []
        if location_keys is None:
            location_keys = []
        if params is None:
            params = {}

        return connector.fetch_pipes_keys(
            connector_keys = connector_keys,
            metric_keys = metric_keys,
            location_keys = location_keys,
            params = params,
            debug = debug
        )

    def _explicit(
            connector_keys : Optional[List[str]] = None,
            metric_keys : Optional[List[str]] = None,
            location_keys : Optional[List[str]] = None,
            params : Optional[Dict[str, Any]] = None,
            debug : bool = False,
            **kw
        ) -> List[Tuple[str, str, str]]:
        """
        Explicitly build Pipes based on provided keys.
        Raises an error if connector_keys or metric_keys is empty,
        and assumes location_keys = [None] if empty.
        """

        if connector_keys is None:
            connector_keys = []
        if metric_keys is None:
            metric_keys = []
        if location_keys is None:
            location_keys = []
        if params is None:
            params = {}

        if not isinstance(connector_keys, list):
            connector_keys = [connector_keys]
        if not isinstance(metric_keys, list):
            metric_keys = [metric_keys]
        if not isinstance(location_keys, list):
            location_keys = [location_keys]

        missing_keys = []
        if len(connector_keys) == 0:
            missing_keys.append('connector_keys')
        if len(metric_keys) == 0:
            missing_keys.append('metric_keys')
        if len(location_keys) == 0:
            location_keys.append(None)
        if len(missing_keys) > 0:
            error_message = "Missing parameters: '" + "', '".join(missing_keys) + "'"
            error_message += "\nSee --help for information for passing parameters."
            error(error_message)
        result = []
        for ck in connector_keys:
            for mk in metric_keys:
                for lk in location_keys:
                    result.append((ck, mk, lk))
        return result

    def _all(**kw):
        """
        Fetch all available metrics and locations and create every combination.
        Connector keys are required.
        """
        error(
            "Need to implement metrics and locations logic in SQL and API.",
            NotImplementedError
        )

    _methods = {
        'registered' : _registered,
        'explicit'   : _explicit,
        'all'        : _all,
        ### TODO implement 'all'
    }
    if method not in _methods:
        error(f"Method '{method}' is not supported!", NotImplementedError)
    return _methods[method](**kw)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Declare FastAPI events in this module (startup, shutdown, etc.).
"""

import sys, os, time
from meerschaum.api import app, get_api_connector, get_uvicorn_config, debug
from meerschaum.utils.debug import dprint
from meerschaum.utils.misc import retry_connect
from meerschaum.config._paths import API_UVICORN_CONFIG_PATH

@app.on_event("startup")
async def startup():
    conn = get_api_connector()
    try:
        connected = retry_connect(
            get_api_connector(),
            workers = get_uvicorn_config().get('workers', None),
            debug = debug
        )
    except Exception as e:
        print(e)
        connected = False
    if not connected:
        await shutdown()
        os._exit(1)

@app.on_event("shutdown")
async def shutdown():
    try:
        if API_UVICORN_CONFIG_PATH.exists() and not debug:
            os.remove(API_UVICORN_CONFIG_PATH)
    except Exception as e:
        pass
        print(e)
    if debug:
        dprint("Closing connection...")
    if get_api_connector().type == 'sql':
        try:
            await get_api_connector().db.disconnect()
        except Exception as e:
            pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Formatting functions for printing pipes
"""

from __future__ import annotations
from meerschaum.utils.typing import PipesDict, Dict

def pprint_pipes(pipes : PipesDict) -> None:
    """
    Print a stylized tree of a Pipes dictionary.
    Supports ANSI and UNICODE global settings.

    This code is pretty unreadable. Just a warning. But it's thoroughly tested,
    so things *shouldn't* break.
    """
    from meerschaum.utils.warnings import error
    from meerschaum.utils.packages import attempt_import, import_rich
    from meerschaum.utils.misc import sorted_dict, replace_pipes_in_dict
    from meerschaum.utils.formatting import UNICODE, ANSI, pprint, colored, get_console
    from meerschaum.config import get_config
    import copy
    rich = import_rich('rich', warn=False)
    Text = None
    if rich is not None:
        rich_text = attempt_import('rich.text')
        Text = rich_text.Text

    icons = {'connector' : '', 'metric' : '', 'location' : '', 'key' : ''}
    styles = {'connector' : '', 'metric' : '', 'location' : '', 'key' : ''}
    guide_style, none_style = '', ''
    if UNICODE:
        icons['connector'] = get_config('formatting', 'emoji', 'connector', patch=True) + ' '
        icons['metric'] = get_config('formatting', 'emoji', 'metric', patch=True) + ' '
        icons['location'] = get_config('formatting', 'emoji', 'location', patch=True) + ' '
        icons['key'] = get_config('formatting', 'emoji', 'key', patch=True) + ' '
    if ANSI:
        styles['connector'] = 'green'
        styles['metric'] = 'bright_blue'
        styles['location'] = 'magenta'
        guide_style = 'dim'
        none_style = 'black on magenta'

    print()

    def ascii_print_pipes():
        """
        Print the dictionary with no unicode allowed. Also works in case rich fails to import
        (though rich should auto-install when `attempt_import()` is called).
        """
        asciitree = attempt_import('asciitree')
        def _replace_pipe_ascii_tree(pipe):
            return {str(pipe) : {}}
        ascii_dict, replace_dict = {}, {'connector' : {}, 'metric' : {}, 'location' : {}}
        for conn_keys, metrics in pipes.items():
            _colored_conn_key = colored(icons['connector'] + conn_keys, style=styles['connector'])
            if Text is not None:
                replace_dict['connector'][_colored_conn_key] = (
                    Text(conn_keys, style=styles['connector'])
                )
            ascii_dict[_colored_conn_key] = {}
            for metric, locations in metrics.items():
                _colored_metric_key = colored(icons['metric'] + metric, style=styles['metric'])
                if Text is not None:
                    replace_dict['metric'][_colored_metric_key] = (
                        Text(metric, style=styles['metric'])
                    )
                ascii_dict[_colored_conn_key][_colored_metric_key] = {}
                for location, pipe in locations.items():
                    if location is None:
                        _location_style = none_style
                    else:
                        _location_style = styles['location']
                    pipe_addendum = '\n         ' + str(pipe)
                    _colored_location = colored(
                        icons['location'] + str(location), style=_location_style
                    )
                    _colored_location_key = _colored_location + pipe_addendum
                    if Text is not None:
                        replace_dict['location'][_colored_location] = (
                            Text(str(location), style=_location_style)
                        )
                    ascii_dict[_colored_conn_key][_colored_metric_key][_colored_location_key] = {}

        tree = asciitree.LeftAligned()
        output = ''
        cols = []

        ### This is pretty terrible, unreadable code.
        ### Please know that I'm normally better than this.
        key_str = (
            (Text("     ") if Text is not None else "     ") +
            (
                Text("Key", style='underline') if Text is not None else
                colored("Key", style='underline')
            ) + (Text('\n\n  ') if Text is not None else '\n\n  ') +
            (
                Text("Connector", style=styles['connector']) if Text is not None else
                colored("Connector", style=styles['connector'])
            ) + (Text('\n   +-- ') if Text is not None else '\n   +-- ') +
            (
                Text("Metric", style=styles['metric']) if Text is not None else
                colored("Metric", style=styles['metric'])
            ) + (Text('\n       +-- ') if Text is not None else '\n       +-- ') +
            (
                Text("Location", style=styles['location']) if Text is not None else
                colored("Location", style=styles['location'])
            ) + (Text('\n\n') if Text is not None else '\n\n')
        )

        output += str(key_str)
        cols.append(key_str)

        def replace_tree_text(tree_str : str) -> Text:
            """
            Replace the colored words with stylized Text instead.
            Is not executed if ANSI and UNICODE are disabled.
            """
            tree_text = Text(tree_str) if Text is not None else None
            for k, v in replace_dict.items():
                for _colored, _text in v.items():
                    parts = []
                    lines = tree_text.split(_colored)
                    for part in lines:
                        parts += [part, _text]
                    if lines[-1] != Text(''):
                        parts = parts[:-1]
                    _tree_text = Text('')
                    for part in parts:
                        _tree_text += part
                    tree_text = _tree_text
            return tree_text

        tree_output = ""
        for k, v in ascii_dict.items():
            branch = {k : v}
            tree_output += tree(branch) + '\n\n'
            if not UNICODE and not ANSI:
                _col = (Text(tree(branch)) if Text is not None else tree(branch))
            else:
                _col = replace_tree_text(tree(branch))
            cols.append(_col)
        if len(output) > 0:
            tree_output = tree_output[:-2]
        output += tree_output

        if rich is None:
            return print(output)

        rich_columns = attempt_import('rich.columns')
        Columns = rich_columns.Columns
        columns = Columns(cols)
        #  rich.print(columns)
        get_console().print(columns)

    if not UNICODE:
        return ascii_print_pipes()

    rich_panel, rich_tree, rich_text, rich_columns, rich_table = attempt_import(
        'rich.panel',
        'rich.tree',
        'rich.text',
        'rich.columns',
        'rich.table',
    )
    from rich import box
    Panel = rich_panel.Panel
    Tree = rich_tree.Tree
    Text = rich_text.Text
    Columns = rich_columns.Columns
    Table = rich_table.Table

    key_panel = Panel(
        (
            Text("\n") +
            Text(icons['connector'] + "Connector", style=styles['connector']) + Text("\n\n") +
            Text(icons['metric'] + "Metric", style=styles['metric']) + Text("\n\n") +
            Text(icons['location'] + "Location", style=styles['location']) + Text("\n")
        ),
        title = Text(icons['key'] + "Keys", style=guide_style),
        border_style = guide_style,
        expand = True
    )

    cols = []
    conn_trees = {}
    metric_trees = {}
    pipes = sorted_dict(pipes)
    for conn_keys, metrics in pipes.items():
        conn_trees[conn_keys] = Tree(
            Text(
                icons['connector'] + conn_keys,
                style = styles['connector'],
            ),
            guide_style = styles['connector']
        )
        metric_trees[conn_keys] = dict()
        for metric, locations in metrics.items():
            metric_trees[conn_keys][metric] = Tree(
                Text(
                    icons['metric'] + metric,
                    style = styles['metric']
                ),
                guide_style = styles['metric']
            )
            conn_trees[conn_keys].add(metric_trees[conn_keys][metric])
            for location, pipe in locations.items():
                if location is None:
                    _location = Text(str(location), style=none_style)
                else:
                    _location = Text(location, style=styles['location'])
                _location = Text(icons['location']) + _location + Text("\n" + str(pipe) + "\n")
                metric_trees[conn_keys][metric].add(_location)

    cols += [key_panel]
    for k, t in conn_trees.items():
        cols.append(t)

    columns = Columns(cols)
    #  rich.print(columns)
    get_console().print(columns)

def pprint_pipe_columns(
        pipe : meerschaum.Pipe,
        nopretty : bool = False,
        debug : bool = False,
    ) -> None:
    """
    Pretty-print a pipe's columns.
    """
    import json
    from meerschaum.utils.warnings import info
    from meerschaum.utils.formatting import pprint, print_tuple, get_console
    from meerschaum.utils.formatting._shell import make_header
    from meerschaum.utils.packages import attempt_import, import_rich

    _cols = pipe.columns
    _cols_types = pipe.get_columns_types(debug=debug)

    def _nopretty_print():
        print(json.dumps(pipe.__getstate__()))
        print(json.dumps(_cols))
        print(json.dumps(_cols_types))

    def _pretty_print():
        rich = import_rich()
        rich_table = attempt_import('rich.table')

        table = rich_table.Table(title=f"Column types for pipe '{pipe}'")
        table.add_column('Column')
        table.add_column('Type', justify='right')

        info(make_header(f"\nColumns for pipe '{pipe}':"), icon=False)
        if _cols:
            pprint(_cols, nopretty=nopretty)
            print()
        else:
            print_tuple((False, f"No registered columns for pipe '{pipe}'."))

        for c, t in _cols_types.items():
            table.add_row(c, t)

        if _cols_types:
            #  rich.print(table)
            get_console().print(table)
            #  print(f"\nTable columns and types:")
            #  pprint(_cols_types, nopretty=nopretty)
        else:
            print_tuple((False, f"No table columns for pipe '{pipe}'. Does the pipe exist?"))

    if nopretty:
        _nopretty_print()
    else:
        _pretty_print()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Start subsystems (API server, logging daemon, etc.).
"""

from __future__ import annotations
from meerschaum.utils.typing import SuccessTuple, Optional, List, Any

def start(
        action : Optional[List[str]] = None,
        **kw : Any,
    ) -> SuccessTuple:
    """
    Start subsystems (API server, logging daemon, etc.).
    """

    from meerschaum.utils.misc import choose_subaction
    options = {
        'api' : _start_api,
        'jobs' : _start_jobs,
    }
    return choose_subaction(action, options, **kw)

def _complete_start(
        action : Optional[List[str]] = None,
        **kw : Any
    ) -> List[str]:
    """
    Override the default Meerschaum `complete_` function.
    """

    if action is None:
        action = []

    options = {
        'job' : _complete_start_jobs,
        'jobs' : _complete_start_jobs,
    }

    if (
        len(action) > 0 and action[0] in options
            and kw.get('line', '').split(' ')[-1] != action[0]
    ):
        sub = action[0]
        del action[0]
        return options[sub](action=action, **kw)

    from meerschaum.actions.shell import default_action_completer
    return default_action_completer(action=(['start'] + action), **kw)

def _start_api(action : Optional[List[str]] = None, **kw):
    """
    Start the API server.

    Usage:
        `start api {options}`

    Options:
        - `-p, --port {number}`
            Port to bind the API server to.

        - `-w, --workers {number}`
            How many worker threads to run.
            Defaults to the number of CPU cores or 1 on Android.
    """
    from meerschaum.actions import actions
    return actions['api'](action=['start'], **kw)

def _start_jobs(
        action : Optional[List[str]] = None,
        name : Optional[str] = None,
        **kw
    ) -> SuccessTuple:
    """
    Run a Meerschaum action as a background job.

    To create a new job, pass the command arguments after `start job`.
    To start a stopped job, pass the job name after `start job`.

    You may also run a background job with the `-d` or `--daemon` flags.

    Examples:

        Create new jobs:

          - `start job sync pipes --loop`
                Run the action `sync pipes --loop` as a background job.
                Generates a random name; e.g. 'happy_seal'.
          
          - `start api --daemon --name api_server`
                Run the action `start api` as a background job, and assign the job
                the name 'api_server'.

        Start stopped jobs:

          - `start job happy_seal`
                Start the job 'happy_seal'.

          - `start job --name happy_seal`
                Start the job 'happy_seal' but via the `--name` flag.
                This only applies when no text follows the words 'start job'.
    """
    import textwrap
    from meerschaum.utils.warnings import warn, info
    from meerschaum.utils.daemon import (
        daemon_action, Daemon, get_daemon_ids, get_daemons, get_filtered_daemons,
        get_stopped_daemons, get_running_daemons
    )
    from meerschaum.utils.daemon._names import get_new_daemon_name
    from meerschaum.actions.arguments._parse_arguments import parse_arguments
    from meerschaum.actions import actions
    from meerschaum.utils.prompt import yes_no
    from meerschaum.utils.formatting import print_tuple
    from meerschaum.utils.formatting._jobs import pprint_job, pprint_jobs
    from meerschaum.utils.formatting._shell import clear_screen
    from meerschaum.utils.misc import items_str

    names = []
    daemon_ids = get_daemon_ids()

    new_job = len(list(action)) > 0
    _potential_jobs = {'known' : [], 'unknown' : []}

    if action:
        for a in action:
            _potential_jobs[('known' if a in daemon_ids else 'unknown')].append(a)
        
        ### Check if the job is named after an action.
        if (
            _potential_jobs['known']
                and _potential_jobs['unknown']
                and _potential_jobs['known'][0] == action[0]
                and _potential_jobs['known'][0] in actions
        ):
            _potential_jobs['unknown'].insert(0, _potential_jobs['known'][0])
            del _potential_jobs['known'][0]

        ### Only spawn a new job if we don't don't find any jobs.
        new_job = (len(_potential_jobs['known']) == 0)
        if not new_job and _potential_jobs['unknown']:
            if not kw.get('nopretty', False):
                warn(
                    (
                        "Unknown job" + ("s" if len(_potential_jobs['unknown']) > 1 else '') + " "
                        + items_str(_potential_jobs['unknown'])
                        + " will be ignored."
                    ),
                    stack = False
                )

        ### Determine the `names` list.
        if new_job:
            names = [(get_new_daemon_name() if not name else name)]
        elif not new_job and not name:
            names = _potential_jobs['known']
        ### Cannot find dameon_id
        else:
            msg = (
                f"Unknown job" + ('s' if len(action) != 1 else '') + ' '
                + items_str(action, and_str='or') + '.'
            )
            return False, msg

    ### No action provided but a --name was. Start job if possible.
    ### E.g. `start job --myjob`
    elif name is not None:
        new_job = False
        names = [name]

    ### No action or --name was provided. Ask to start all stopped jobs.
    else:
        _stopped_daemons = get_stopped_daemons()
        if not _stopped_daemons:
            return False, "No jobs to start."
        names = [d.daemon_id for d in _stopped_daemons]

    def _run_new_job(name : Optional[str] = None):
        kw['action'] = action
        if not name:
            name = get_new_daemon_name()
        kw['name'] = name
        _action_success_tuple = daemon_action(daemon_id=name, **kw)
        return _action_success_tuple, name

    def _run_existing_job(name : Optional[str] = None):
        daemon = Daemon(daemon_id=name)

        if not daemon.path.exists():
            if not kw.get('nopretty', False):
                warn(f"There isn't a job with the name '{name}'.", stack=False)
                print(
                    f"You can start a new job named '{name}' with `start job "
                    + "{options}" + f" --name {name}`"
                )
            return (False, f"Job '{name}' does not exist."), daemon.daemon_id

        daemon.cleanup()
        try:
            _daemon_sysargs = daemon.properties['target']['args'][0]
        except KeyError:
            return False, "Failed to get arguments for daemon '{dameon.daemon_id}'."
        _daemon_kw = parse_arguments(_daemon_sysargs)
        _daemon_kw['name'] = daemon.daemon_id
        _action_success_tuple = daemon_action(
            **_daemon_kw
        )
        if not _action_success_tuple[0]:
            return _action_success_tuple, daemon.daemon_id
        return (True, f"Success"), daemon.daemon_id

    if not names:
        return False, "No jobs to start."

    ### Get user permission to clear logs.
    _filtered_daemons = get_filtered_daemons(names)
    if not kw.get('force', False) and _filtered_daemons:
        _filtered_running_daemons = get_running_daemons(_filtered_daemons)
        if _filtered_running_daemons:
            pprint_jobs(_filtered_running_daemons)
            if yes_no(
                "The above jobs are still running. Do you want to first stop these jobs?",
                default = 'n',
                yes = kw.get('yes', False),
                noask = kw.get('noask', False)
            ):
                stop_success_tuple = actions['stop'](
                    action = ['jobs'] + [d.daemon_id for d in _filtered_running_daemons],
                    force = True,
                )
                if not stop_success_tuple[0]:
                    warn(
                        "Failed to stop job" + ("s" if len(_filtered_running_daemons) != 1 else '')
                        + items_str([d.daemon_id for d in _filtered_running_daemons])
                        + ".",
                        stack = False
                    )
                    for d in _filtered_running_daemons:
                        names.remove(d.daemon_id)
                        _filtered_daemons.remove(d)
            else:
                info(
                    "Skipping already running job"
                    + ("s" if len(_filtered_running_daemons) != 1 else '') + ' '
                    + items_str([d.daemon_id for d in _filtered_running_daemons]) + '.'
                )
                for d in _filtered_running_daemons:
                    names.remove(d.daemon_id)
                    _filtered_daemons.remove(d)

        if not _filtered_daemons:
            return False, "No jobs to start."
        pprint_jobs(_filtered_daemons, nopretty=kw.get('nopretty', False))
        if not yes_no(
            (
                f"Would you like to overwrite the logs and run the job"
                + ("s" if len(names) != 1 else '') + " " + items_str(names) + "?"
            ),
            default = 'n',
            yes = kw.get('yes', False),
            nopretty = kw.get('nopretty', False),
            noask = kw.get('noask', False),
        ):
            return (False, "Nothing was started.")


    _successes, _failures = [], []
    for _name in names:
        success_tuple, __name = _run_new_job(_name) if new_job else _run_existing_job(_name)
        if success_tuple[0]:
            if kw.get('nopretty', False):
                print_tuple(True, f"Successfully started job '{__name}'.")
            _successes.append(_name)
        else:
            _failures.append(_name)

    msg = (
        (("Successfully started job" + ("s" if len(_successes) != 1 else '')
            + f" {items_str(_successes)}." + ('\n' if _failures else ''))
            if _successes else '')
        + ("Failed to start job" + ("s" if len(_failures) != 1 else '')
            + f" {items_str(_failures)}." if _failures else '')
    )
    return len(_successes) > 0, msg

def _complete_start_jobs(
        action : Optional[List[str]] = None,
        line : str = '',
        **kw
    ) -> List[str]:
    from meerschaum.utils.daemon import get_daemon_ids
    daemon_ids = get_daemon_ids()
    if not action:
        return daemon_ids
    possibilities = []
    #  if action[-1] in daemon_ids:
        #  return daemon_ids
    _line_end = line.split(' ')[-1]
    for daemon_id in daemon_ids:
        #  if daemon_id.startswith(action[-1]) and (
            #  daemon_id not in action or _line_end == ''
        #  ):
        if daemon_id in action:
            continue
        if _line_end == '':
            possibilities.append(daemon_id)
            continue
        if daemon_id.startswith(action[-1]):
            possibilities.append(daemon_id)
    return possibilities


### NOTE: This must be the final statement of the module.
###       Any subactions added below these lines will not
###       be added to the `help` docstring.
from meerschaum.utils.misc import choices_docstring as _choices_docstring
start.__doc__ += _choices_docstring('start')



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Define file paths
"""

from __future__ import annotations

from pathlib import Path
import os, platform
try:
    import importlib.resources as pkg_resources
except ImportError:
    import importlib_resources as pkg_resources

package_root_context_manager = pkg_resources.path('meerschaum', '__init__.py')
with package_root_context_manager as file_path:
    PACKAGE_ROOT_PATH = Path(os.path.join(Path(file_path.parent.parent), 'meerschaum'))

paths = {
    'PACKAGE_ROOT_PATH' : PACKAGE_ROOT_PATH,
    'ROOT_DIR_PATH' : (
        Path(os.path.join(Path.home(), '.config', 'meerschaum'))
        if platform.system() != 'Windows'
        else Path(os.path.join(os.environ['AppData'], 'Meerschaum'))
    ),
    'VIRTENV_RESOURCES_PATH' : ('{ROOT_DIR_PATH}', 'venvs'),
    'CONFIG_DIR_PATH' : ('{ROOT_DIR_PATH}', 'config'),
    'DEFAULT_CONFIG_DIR_PATH' : ('{ROOT_DIR_PATH}', 'default_config'),
    'PATCH_DIR_PATH' : ('{ROOT_DIR_PATH}', 'patch_config'),
    'PERMANENT_PATCH_DIR_PATH' : ('{ROOT_DIR_PATH}', 'permanent_patch_config'),

    'STACK_RESOURCES_PATH' : ('{ROOT_DIR_PATH}', 'stack'),
    'STACK_COMPOSE_FILENAME' : 'docker-compose.yaml',
    'STACK_COMPOSE_PATH' : ('{STACK_RESOURCES_PATH}', '{STACK_COMPOSE_FILENAME}'),
    'STACK_ENV_FILENAME' : '.env',
    'STACK_ENV_PATH' : ('{STACK_RESOURCES_PATH}', '{STACK_ENV_FILENAME}'),

    'SHELL_RESOURCES_PATH' : ('{ROOT_DIR_PATH}', ),
    'SHELL_HISTORY_PATH' : ('{SHELL_RESOURCES_PATH}', '.mrsm_history'),

    'API_RESOURCES_PATH' : ('{PACKAGE_ROOT_PATH}', 'api', 'resources'),
    'API_STATIC_PATH' : ('{API_RESOURCES_PATH}', 'static'),
    'API_TEMPLATES_PATH' : ('{API_RESOURCES_PATH}', 'templates'),
    'API_CONFIG_RESOURCES_PATH' : ('{ROOT_DIR_PATH}', 'api'),
    'API_SECRET_KEY_PATH' : ('{API_CONFIG_RESOURCES_PATH}', '.api_secret_key'),
    'API_UVICORN_RESOURCES_PATH' : ('{API_CONFIG_RESOURCES_PATH}', 'uvicorn'),
    'API_UVICORN_CONFIG_PATH' : ('{API_UVICORN_RESOURCES_PATH}', '.thread_config.json'),

    'CACHE_RESOURCES_PATH' : ('{ROOT_DIR_PATH}', '.cache'),
    'PIPES_CACHE_RESOURCES_PATH' : ('{ROOT_DIR_PATH}', 'pipes'),
    'USERS_CACHE_RESOURCES_PATH' : ('{ROOT_DIR_PATH}', 'users'),

    'PLUGINS_RESOURCES_PATH' : ('{ROOT_DIR_PATH}', 'plugins'),
    'PLUGINS_ARCHIVES_RESOURCES_PATH' : ('{PLUGINS_RESOURCES_PATH}', '.archives'),
    'PLUGINS_TEMP_RESOURCES_PATH' : ('{PLUGINS_RESOURCES_PATH}', '.tmp'),
    'PLUGINS_INIT_PATH' : ('{PLUGINS_RESOURCES_PATH}', '__init__.py'),

    'SQLITE_RESOURCES_PATH' : ('{ROOT_DIR_PATH}', 'sqlite'),
    'SQLITE_DB_PATH' : ('{SQLITE_RESOURCES_PATH}', 'mrsm_local.db'),

    'DUCKDB_RESOURCES_PATH' : ('{ROOT_DIR_PATH}', 'duckdb'),
    'DUCKDB_PATH' : ('{DUCKDB_RESOURCES_PATH}', 'duck.db'),

    'GRAFANA_RESOURCES_PATH' : ('{STACK_RESOURCES_PATH}', 'grafana', 'resources'),
    'GRAFANA_DATASOURCE_PATH' : (
        '{GRAFANA_RESOURCES_PATH}', 'provisioning', 'datasources', 'datasource.yaml'
    ),
    'GRAFANA_DASHBOARD_PATH' : (
        '{GRAFANA_RESOURCES_PATH}', 'provisioning', 'dashboards', 'dashboard.yaml'
    ),
    'MOSQUITTO_RESOURCES_PATH' : ('{STACK_RESOURCES_PATH}', 'mosquitto', 'resources'),
    'MOSQUITTO_CONFIG_PATH' : ('{MOSQUITTO_RESOURCES_PATH}', 'mosquitto.conf'),

    'PORTABLE_CHECK_READLINE_PATH' : ('{SHELL_RESOURCES_PATH}', '.readline_attempted_install'),

    'DAEMON_RESOURCES_PATH' : ('{ROOT_DIR_PATH}', 'jobs'),
    'LOGS_RESOURCES_PATH' : ('{ROOT_DIR_PATH}', 'logs'),
}

def set_root(root : Union[Path, str]):
    """
    Modify the value of `ROOT_DIR_PATH`.
    """
    global paths
    paths['ROOT_DIR_PATH'] = Path(root)

def __getattr__(name : str) -> Path:
    if name not in paths:
        if name not in globals():
            raise AttributeError(f"Could not import '{name}'.")
        return globals()[name]

    if isinstance(paths[name], (list, tuple)):
        ### recurse through paths to create resource directories.
        parts = []
        for p in paths[name]:
            if str(p).startswith('{') and str(p).endswith('}'):
                parts.append(__getattr__(p[1:-1]))
                #  parts.append(paths[p[1:-1]])
            else:
                parts.append(p)
        path = Path(os.path.join(*parts))
    else:
        path = Path(paths[name])

    ### Create directories or touch files.
    if 'RESOURCES_PATH' in name or name == 'CONFIG_DIR_PATH':
        path.mkdir(parents=True, exist_ok=True)
    elif 'FILENAME' in name:
        path = str(path)

    return path




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
The PluginConnector imports fetch and sync methods from a Plugin (if possible) and
allows Pipes to have connectors of type 'plugin'.
"""

from meerschaum.connectors import Connector

class PluginConnector(Connector):

    def __init__(
        self,
        label : str,
        debug : bool = False,
        **kw
    ):
        """
        The PluginConnector imports fetch and sync methods from a Plugin (if possible) and
        allows Pipes to have connectors of type 'plugin'.
        """
        super().__init__('plugin', label=label, **kw)

        import os, pathlib, sys
        from meerschaum.utils.warnings import error, warn
        from meerschaum.config._paths import PLUGINS_RESOURCES_PATH
        if PLUGINS_RESOURCES_PATH not in sys.path:
            sys.path.append(str(PLUGINS_RESOURCES_PATH))

        self.resource_path = None
        for _plugin in os.listdir(PLUGINS_RESOURCES_PATH):
            plugin = _plugin.replace('.py', '')
            if plugin == self.label:
                self.resource_path = pathlib.Path(os.path.join(PLUGINS_RESOURCES_PATH, plugin))
                break
        if not self.resource_path:
            error(f"Plugin '{self.label}' cannot be found. Is it installed?")

        self.fetch = None
        try:
            exec(f'from plugins.{self.label} import fetch; self.fetch = fetch')
        except Exception as e:
            pass

        self.sync = None
        try:
            exec(f'from plugins.{self.label} import sync; self.sync = sync')
        except Exception as e:
            pass

        if self.fetch is None and self.sync is None:
            error(f"Could not import `fetch()` or `sync()` methods for plugin '{self.label}'")



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Launch into a CLI environment to interact with the SQL Connector
"""

flavor_clis = {
    'postgresql'  : 'pgcli',
    'timescaledb' : 'pgcli',
    'cockroachdb' : 'pgcli',
    'mysql'       : 'mycli',
    'mariadb'     : 'mycli',
    'percona'     : 'mycli',
    'sqlite'      : 'litecli',
    'mssql'       : 'mssqlcli', ### NOTE: Not added as a dependency due to dependency problems
}

def cli(
        self,
        debug : bool = False
    ):
    """
    Launch an interactive CLI for the SQLConnector's flavor
    """
    from meerschaum.utils.packages import venv_exec, attempt_import
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.warnings import error
    import sys, subprocess

    if self.flavor not in flavor_clis:
        return False, f"No CLI available for flavor '{self.flavor}'."

    ### get CLI module to launch
    cli_name = flavor_clis[self.flavor]
    if debug:
        dprint(f"Opening CLI '{cli_name}' for {self} (flavor '{self.flavor}')...")

    ### attempt an import to raise warnings if not installed
    cli = attempt_import(cli_name, lazy=False, debug=debug)

    ### open sqlalchemy engine URI or just database if sqlite
    cli_arg_str = self.DATABASE_URL
    if self.flavor == 'sqlite':
        cli_arg_str = str(self.database)

    ### run the module in a subprocess because it calls sys.exit(), and __main__ does not
    ### work for these CLIs (something to do with Click?)
    launch_cli = f"from {cli_name} import main; main.cli(['{cli_arg_str}'])"

    success = venv_exec(launch_cli, debug=debug)
    return success, f"CLI exited with response: {success}"



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Insert non-user-editable configuration files here.
"""

static_config = None

def _static_config():
    global static_config
    if static_config is None:
        static_config = {
            'api' : {
                'endpoints' : {
                    'index' : '/',
                    'favicon' : '/favicon.ico',
                    'plugins' : '/plugins',
                    'pipes' : '/pipes',
                    'metadata' : '/metadata',
                    'actions' : '/actions',
                    'users' : '/users',
                    'login' : '/login',
                    'connectors' : '/connectors',
                    'version' : '/version',
                    'chaining' : '/chaining',
                    'websocket' : '/ws',
                    'dash' : '/dash',
                },
                'oauth' : {
                    'token_expires_minutes' : 15,
                },
            },
            'environment' : {
                'config' : 'MRSM_CONFIG',
                'patch' : 'MRSM_PATCH',
                'root' : 'MRSM_ROOT_DIR',
                'runtime' : 'MRSM_RUNTIME',
            },
            'config' : {
                'default_filetype' : 'json',
                'symlinks_key' : '_symlinks',
            },
            'system' : {
                'arguments' : {
                    'sub_decorators' : (
                        '[',
                        ']'
                    ),
                },
                'urls' : {
                    'get-pip.py' : 'https://bootstrap.pypa.io/get-pip.py',
                },
                'success' : {
                    'ignore' : (
                        'Success',
                        'success'
                        'Succeeded',
                        '',
                        None,
                    ),
                },
                'prompt' : {
                    'web' : False,
                },
            },
            'connectors' : {
                'default_label' : 'main',
            },
            'users' : {
                'password_hash' : {
                    'schemes' : [
                        'pbkdf2_sha256',
                    ],
                    'default' : 'pbkdf2_sha256',
                    'pbkdf2_sha256__default_rounds' : 30000,
                },
            },
        }
    return static_config



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
"""
Miscellaneous functions go here
"""

from __future__ import annotations
from meerschaum.utils.typing import (
    Union, Mapping, Any, Callable, Optional, List, Dict, SuccessTuple, Iterable, PipesDict
)

def add_method_to_class(
        func : Callable[[Any], Any],
        class_def : 'Class',
        method_name : Optional[str] = None,
        keep_self : Optional[bool] = None,
    ) -> Callable[[Any], Any]:
    """
    Add function `func` to class `class_def`.

    :param func:
        Function to be added as a method of the class

    :param class_def:
        Class we are modifying

    :param method_name:
        New name of the method. None will use func.__name__ (default).
    """
    from functools import wraps

    is_class = isinstance(class_def, type)
    
    @wraps(func)
    def wrapper(self, *args, **kw):
        print(self, args, kw)
        return func(*args, **kw)

    if method_name is None:
        method_name = func.__name__

    setattr(class_def, method_name, (
            wrapper if ((is_class and keep_self is None) or keep_self is False) else func
        )
    )

    return func

def choose_subaction(
        action : Optional[List[str]] = None,
        options : Optional[Dict[str, Any]] = None,
        **kw
    ) -> SuccessTuple:
    """
    Given a dictionary of options and the standard Meerschaum actions list,
    check if choice is valid and execute chosen function, else show available
    options and return False

    :param action:
        subactions (e.g. `show pipes` -> ['pipes'])
    
    :param options:
        Available options to execute.
        option (key) -> function (value)
        Functions must accept **kw keyword arguments
        and return a tuple of success bool and message.
    """
    from meerschaum.utils.warnings import warn, info
    import inspect
    if action is None:
        action = []
    if options is None:
        options = {}
    parent_action = inspect.stack()[1][3]
    if len(action) == 0:
        action = ['']
    choice = action[0]

    def valid_choice(_choice : str, _options : dict):
        if _choice in _options:
            return _choice
        if (_choice + 's') in options:
            return _choice + 's'
        return None

    parsed_choice = valid_choice(choice, options)
    if parsed_choice is None:
        warn(f"Cannot {parent_action} '{choice}'. Choose one:", stack=False)
        for option in sorted(options):
            print(f"  - {parent_action} {option}")
        return (False, f"Invalid choice '{choice}'.")
    ### remove parent sub-action
    kw['action'] = list(action)
    del kw['action'][0]
    return options[parsed_choice](**kw)

def generate_password(
        length : int = 12
    ) -> str:
    """
    Generate a secure password of given length.
    """
    import secrets, string
    return ''.join((secrets.choice(string.ascii_letters) for i in range(length)))

def is_int(s : str) -> bool:
    """
    Check if string is an int.
    """
    try:
        float(s)
    except Exception as e:
        return False
    
    return float(s).is_integer()

def get_options_functions():
    """
    Get options functions from parent module
    """
    import inspect
    parent_globals = inspect.stack()[1][0].f_globals
    parent_package = parent_globals['__name__']
    print(parent_package)

def string_to_dict(
        params_string : str
    ) -> Dict[str, Any]:
    """
    Parse a string into a dictionary

    If the string begins with '{', parse as JSON. Else use simple parsing

    """
    if params_string == "":
        return dict()

    if str(params_string).startswith('{'):
        import json
        return json.loads(params_string)

    import ast
    params_dict = dict()
    for param in params_string.split(","):
        _keys = param.split(":")
        keys = _keys[:-1]
        try:
            val = _keys[-1]
        except Exception as e:
            val = str(_keys[-1])

        c = params_dict
        for _k in keys[:-1]:
            try:
                k = ast.literal_eval(_k)
            except Exception as e:
                k = str(_k)
            if k not in c:
                c[k] = {}
            c = c[k]

        c[keys[-1]] = val

    return params_dict

def parse_config_substitution(
        value : str,
        leading_key : str = 'MRSM',
        begin_key : str = '{',
        end_key : str = '}',
        delimeter : str = ':'
    ):
    """
    Parse Meerschaum substitution syntax
    E.g. MRSM{value1:value2} => ['value1', 'value2']
    NOTE: Not currently used. See `search_and_substitute_config` in `meerschaum.config._read_yaml`.
    """
    if not value.beginswith(leading_key):
        return value

    return leading_key[len(leading_key):][len():-1].split(delimeter)

def edit_file(
        path : Union[pathlib.Path, str],
        default_editor : str = 'pyvim',
        debug : bool = False
    ) -> bool:
    """
    Open a file for editing.

    Attempt to launch the user's defined $EDITOR, otherwise use pyvim.
    """
    import os
    from subprocess import call
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.packages import run_python_package, attempt_import, package_venv
    try:
        EDITOR = os.environ.get('EDITOR', default_editor)
        if debug:
            dprint(f"Opening file '{path}' with editor '{EDITOR}'")
        rc = call([EDITOR, path])
    except Exception as e: ### can't open with default editors
        if debug:
            dprint(e)
            dprint('Failed to open file with system editor. Falling back to pyvim...')
        pyvim = attempt_import('pyvim', lazy=False)
        rc = run_python_package('pyvim', [path], venv=package_venv(pyvim), debug=debug)
    return rc == 0

def is_pipe_registered(
        pipe : Union['meerschaum.Pipe', 'meerschaum.Pipe.MetaPipe'],
        pipes : dict,
        debug : bool = False
    ):
    """
    Check if a Pipe or MetaPipe is inside the pipes dictionary.
    """
    from meerschaum.utils.debug import dprint
    ck, mk, lk = pipe.connector_keys, pipe.metric_key, pipe.location_key
    if debug:
        dprint(f'{ck}, {mk}, {lk}')
        dprint(f'{pipe}, {pipes}')
    return ck in pipes and mk in pipes[ck] and lk in pipes[ck][mk]

def _get_subaction_names(action : str, globs : dict = None) -> List[str]:
    """
    NOTE: Don't use this function. You should use `meerschaum.actions.get_subactions()` instead.
    This only exists for internal use.

    Return a list of function pointers to all subactions for a given action.
    """
    if globs is None:
        import importlib
        module = importlib.import_module(f'meerschaum.actions.{action}')
        globs = vars(module)
    subactions = []
    for item in globs:
        if f'_{action}' in item and 'complete' not in item.lstrip('_'):
            subactions.append(globs[item])
    return subactions

def choices_docstring(action : str, globs : Optional[Dict[str, Any]] = None) -> str:
    """
    Append the an action's available options to the module docstring.
    This function is to be placed at the bottom of each action module.
    """
    options_str = f"\n    Options:\n        `{action} "
    subactions = _get_subaction_names(action, globs=globs)
    options_str += "["
    sa_names = []
    for sa in subactions:
        try:
            sa_names.append(sa.__name__[len(f"_{action}") + 1:])
        except Exception as e:
            print(e)
            return ""
    for sa_name in sorted(sa_names):
        options_str += f"{sa_name}, "
    options_str = options_str[:-2] + "]`"
    return options_str

def print_options(
        options : Optional[Dict[str, Any]] = None,
        nopretty : bool = False,
        no_rich : bool = False,
        name : str = 'options',
        header : Optional[str] = None,
        actions : bool = False,
        num_cols : int = 8,
        **kw
    ) -> None:
    """
    Show available options from an iterable
    """
    from meerschaum.utils.packages import import_rich
    from meerschaum.utils.formatting import make_header
    from meerschaum.actions import actions as _actions

    if options is None:
        options = {}
    _options = []
    for o in options:
        _options.append(str(o))
    _header = f"Available {name}" if header is None else header

    def _print_options_no_rich():
        if not nopretty:
            print()
            print(make_header(_header))
        ### print actions
        for option in sorted(_options):
            if not nopretty:
                print("  - ", end="")
            print(option)
        if not nopretty:
            print()

    rich = import_rich()
    if rich is None or nopretty or no_rich:
        _print_options_no_rich()
        return None

    from meerschaum.utils.formatting import pprint, get_console
    from meerschaum.utils.packages import attempt_import
    rich_columns = attempt_import('rich.columns')
    rich_panel = attempt_import('rich.panel')
    rich_table = attempt_import('rich.table')
    from rich import box
    Panel = rich_panel.Panel
    Columns = rich_columns.Columns
    Table = rich_table.Table

    if _header is not None:
        table = Table(
            title = '\n' + _header,
            box = box.SIMPLE,
            show_header = False,
            show_footer = False,
            title_style = ''
        )
    else:
        table = Table.grid(padding=(0, 2))
    for i in range(num_cols):
        table.add_column()

    chunks = iterate_chunks(sorted(_options), num_cols, fillvalue='')
    for c in chunks:
        table.add_row(*c)

    cols = Columns([
        o for o in sorted(_options)
    ], expand=True, equal=True, title=header, padding=(0, 0))
    get_console().print(table)
    return None

def iterate_chunks(iterable, chunksize : int, fillvalue : Optional[Any] = None):
    """
    Iterate over a list in chunks.
    Found here:
        https://stackoverflow.com/questions/
        434287/what-is-the-most-pythonic-way-to-iterate-over-a-list-in-chunks
    """
    from itertools import zip_longest
    args = [iter(iterable)] * chunksize
    return zip_longest(*args, fillvalue=fillvalue)

def sorted_dict(d : dict) -> dict:
    """
    Sort a dictionary's keys and values and return a new dictionary
    """
    try:
        return {key: value for key, value in sorted(d.items(), key=lambda item: item[1])}
    except Exception as e:
        return d

def flatten_pipes_dict(pipes_dict : dict) -> list:
    """
    Convert the standard pipes dictionary into a list
    """
    pipes_list = []
    for ck in pipes_dict.values():
        for mk in ck.values():
            pipes_list += list(mk.values())
    return pipes_list

def round_time(
        dt : 'datetime.datetime' = None,
        date_delta : 'datetime.timedelta' = None,
        to : 'str' = 'down'
    ) -> 'datetime.datetime':
    """
    Round a datetime object to a multiple of a timedelta
    dt : datetime.datetime object, default now.
    dateDelta : timedelta object, we round to a multiple of this, default 1 minute.
    Function found from:
    http://stackoverflow.com/questions/3463930/how-to-round-the-minute-of-a-datetime-object-python
    """
    import datetime
    if date_delta is None:
        date_delta = datetime.timedelta(minutes=1)
    round_to = date_delta.total_seconds()
    if dt is None:
        dt = datetime.datetime.utcnow()
    seconds = (dt.replace(tzinfo=None) - dt.min.replace(tzinfo=None)).seconds

    if seconds % round_to == 0 and dt.microsecond == 0:
        rounding = (seconds + round_to / 2) // round_to * round_to
    else:
        if to == 'up':
            # // is a floor division, not a comment on following line (like in javascript):
            rounding = (seconds + dt.microsecond/1000000 + round_to) // round_to * round_to
        elif to == 'down':
            rounding = seconds // round_to * round_to
        else:
            rounding = (seconds + round_to / 2) // round_to * round_to

    return dt + datetime.timedelta(0, rounding - seconds, - dt.microsecond)

def parse_df_datetimes(
        df : 'pd.DataFrame',
        debug : bool = False
    ) -> 'pd.DataFrame':
    """
    Parse a pandas DataFrame for datetime columns and cast as datetimes
    """
    from meerschaum.utils.packages import import_pandas
    ### import pandas (or pandas replacement)
    from meerschaum.utils.debug import dprint
    pd = import_pandas()

    ### if df is a dict, build DataFrame
    if not isinstance(df, pd.DataFrame):
        if debug:
            dprint(f"df is not a DataFrame. Casting to DataFrame...")
        df = pd.DataFrame(df)

    ### skip parsing if DataFrame is empty
    if len(df) == 0:
        if debug:
            dprint(f"df is empty. Returning original DataFrame without casting datetime columns...")
        return df

    ### apply regex to columns to determine which are ISO datetimes
    iso_dt_regex = r'\d{4}-\d{2}-\d{2}.\d{2}\:\d{2}\:\d{2}'
    dt_mask = df.astype(str).apply(
        lambda s : s.str.match(iso_dt_regex).all()
    )

    ### list of datetime column names
    datetimes = list(df.loc[:, dt_mask])
    if debug:
        dprint("Converting columns to datetimes: " + str(datetimes))

    ### apply to_datetime
    df[datetimes] = df[datetimes].apply(pd.to_datetime)

    ### strip timezone information
    for dt in datetimes:
        df[dt] = df[dt].dt.tz_localize(None)

    return df

def timed_input(
        seconds : int = 10,
        timeout_message : str = "",
        prompt : str = "",
        icon : bool = False,
        **kw
    ) -> Optional[str]:
    """
    Accept user input only for a brief period of time.
    """
    from meerschaum.utils.prompt import prompt as _prompt
    import signal

    class TimeoutExpired(Exception):
        """
        Raise an exception when the timer runs out.
        """

    def alarm_handler(signum, frame):
        raise TimeoutExpired

    # set signal handler
    signal.signal(signal.SIGALRM, alarm_handler)
    signal.alarm(seconds) # produce SIGALRM in `timeout` seconds

    try:
        #  return _prompt(prompt, icon=icon, **kw)
        return input(prompt)
    except TimeoutExpired:
        return None
    finally:
        signal.alarm(0) # cancel alarm

def retry_connect(
        connector : Union[
            meerschaum.connectors.sql.SQLConnector,
            meerschaum.connectors.api.APIConnector,
            None
        ] = None,
        max_retries : int = 40,
        retry_wait : int = 3,
        workers : int = 1,
        warn : bool = True,
        debug : bool = False,
    ):
    """
    Keep trying to connect to the database.
    """
    from meerschaum.utils.warnings import warn as _warn, error, info
    from meerschaum.utils.debug import dprint
    from meerschaum import get_connector
    from meerschaum.utils.packages import attempt_import
    from meerschaum.connectors.sql._tools import test_queries
    from functools import partial
    import time

    ### Get default connector if None is provided.
    if connector is None:
        connector = get_connector()

    if connector.type not in ('sql', 'api'):
        return None

    retries = 0
    connected, chaining_status = False, None
    while retries < max_retries:
        if debug:
            dprint(f"Trying to connect to '{connector}'...")
            dprint(f"Attempt ({retries + 1} / {max_retries})")

        if connector.type == 'sql':

            def _connect(_connector):
                ### Test queries like `SELECT 1`.
                connect_query = test_queries.get(connector.flavor, test_queries['default'])
                if _connector.exec(connect_query) is None:
                    raise Exception("Failed to connect.")

            try:
                _connect(connector)
                connected = True
            except Exception as e:
                print(e)
                connected = False

        elif connector.type == 'api':
            ### If the remote instance does not allow chaining, don't even try logging in.
            if not isinstance(chaining_status, bool):
                chaining_status = connector.get_chaining_status(debug=debug)
                if chaining_status is None:
                    connected = None
                elif chaining_status is False:
                    if warn:
                        _warn(
                            f"Meerschaum instance '{connector}' does not allow chaining " +
                            "and cannot be used as the parent for this instance.",
                            stack = False
                        )
                    return False
            if chaining_status:
                connected = connector.login(debug=debug)[0]
                if not connected and warn:
                    _warn(f"Unable to login to '{connector}'!", stack=False)

        if connected:
            if debug:
                dprint("Connection established!")
            return True

        if warn:
            _warn(
                f"Connection failed. Press [Enter] to retry or wait {retry_wait} seconds.",
                stack = False
            )
            info(
                f"To quit, press CTRL-C, then 'q' + Enter for each worker" +
                (f" ({workers})." if workers is not None else ".")
            )
        try:
            if retry_wait > 0:
                text = timed_input(retry_wait)
                if text in ('q', 'quit', 'pass', 'exit', 'stop'):
                    return None
        except KeyboardInterrupt:
            return None
        retries += 1

    return False

def df_from_literal(
        pipe : Optional['meerschaum.Pipe'] = None,
        literal : str = None,
        debug : bool = False
    ) -> 'pd.DataFrame':
    """
    Parse a literal (if nessary), and use a Pipe's column names to generate a DataFrame.
    """
    from meerschaum.utils.packages import import_pandas
    from meerschaum.utils.warnings import error, warn
    from meerschaum.utils.debug import dprint

    if pipe is None or literal is None:
        error("Please provide a Pipe and a literal value")
    ### this will raise an error if the columns are undefined
    dt_name, val_name = pipe.get_columns('datetime', 'value')

    val = literal
    if isinstance(literal, str):
        if debug:
            dprint(f"Received literal string: '{literal}'")
        import ast
        try:
            val = ast.literal_eval(literal)
        except Exception as e:
            warn(
                "Failed to parse value from string:\n" + f"{literal}" +
                "\n\nWill cast as a string instead."\
            )
            val = literal

    ### NOTE: we do everything in UTC if possible.
    ### In dealing with timezones / Daylight Savings lies madness.
    import datetime
    now = datetime.datetime.utcnow()

    pd = import_pandas()
    return pd.DataFrame({dt_name : [now], val_name : [val]})

def filter_unseen_df(
        old_df : 'pd.DataFrame',
        new_df : 'pd.DataFrame',
        dtypes : dict = None,
        custom_nan : str = 'mrsm_NaN',
        debug : bool = False,
    ) -> 'pd.DataFrame':
    """
    Left join two DataFrames to find the newest unseen data.

    I have scoured the web for the best way to do this.
    My intuition was to join on datetime and id, but the code below accounts for values as well
    without needing to define expicit columns or indices.

    The logic below is based off this StackOverflow question, with an index reset thrown on top:
    https://stackoverflow.com/questions/
    48647534/python-pandas-find-difference-between-two-data-frames#48647840

    Also, NaN apparently does not equal NaN, so I am temporarily replacing instances of NaN with a
    custom string, per this StackOverflow question:
    https://stackoverflow.com/questions/31833635/pandas-checking-for-nan-not-working-using-isin

    Lastly, use the old DataFrame's columns for the new DataFrame,
    because order matters when checking equality.
    """
    if old_df is None:
        return new_df
    old_cols = list(old_df.columns)
    try:
        new_df = new_df[old_cols]
    except Exception as e:
        from meerschaum.utils.warnings import warn
        warn(
            "Was not able to cast old columns onto new DataFrame. " +
            f"Are both DataFrames the same shape? Error:\n{e}"
        )
        return None

    ### assume the old_df knows what it's doing, even if it's technically wrong.
    if dtypes is None:
        dtypes = dict(old_df.dtypes)
    new_df = new_df.astype(dtypes)

    if len(old_df) == 0:
        return new_df

    return new_df[
        ~new_df.fillna(custom_nan).apply(tuple, 1).isin(old_df.fillna(custom_nan).apply(tuple, 1))
    ].reset_index(drop=True)

def change_dict(d : Dict[Any, Any], func : 'function') -> None:
    """
    Originally was local, moving to global for multiprocessing debugging
    """
    for k, v in d.items():
        if isinstance(v, dict):
            change_dict(v, func)
        else:
            d[k] = func(v)

def replace_pipes_in_dict(
        pipes : Optional[PipesDict] = None,
        func : 'function' = str,
        debug : bool = False,
        **kw
    ) -> PipesDict:
    """
    Replace the Pipes in a Pipes dict with the result of another function
    """
    if pipes is None:
        from meerschaum import get_pipes
        pipes = get_pipes(debug=debug, **kw)

    result = pipes.copy()
    change_dict(result, func)
    return result

def enforce_gevent_monkey_patch():
    """
    Check if gevent monkey patching is enabled, and if not, then apply patching
    """
    from meerschaum.utils.packages import attempt_import
    import socket
    gevent, gevent_socket, gevent_monkey = attempt_import(
        'gevent', 'gevent.socket', 'gevent.monkey'
    )
    if not socket.socket is gevent_socket.socket:
        gevent_monkey.patch_all()

def is_valid_email(email : str) -> bool:
    """
    Check whether a string is a valid email
    """
    import re
    regex = '^[a-z0-9]+[\._]?[a-z0-9]+[@]\w+[.]\w{2,3}$'
    return re.search(regex, email)

def string_width(string : str) -> int:
    """
    Calculate the width of a string after its last newline
    """
    found_newline = False
    width = 0
    for c in reversed(string):
        if c == '\n' and found_newline:
            break
        if c == '\n':
            found_newline = True
            continue
        if found_newline:
            width += 1
    return width

def _pyinstaller_traverse_dir(
        directory : str,
        ignore_patterns : Iterable[str] = ('.pyc', 'dist', 'build', '.git', '.log'),
        include_dotfiles : bool = False
    ) -> list:
    """
    Recursively traverse a directory and return a list of its contents.
    """
    import os, pathlib
    paths = []
    _directory = pathlib.Path(directory)

    def _found_pattern(name : str):
        for pattern in ignore_patterns:
            if pattern.replace('/', os.path.sep) in str(name):
                return True
        return False

    for root, dirs, files in os.walk(_directory):
        _root = str(root)[len(str(_directory.parent)):]
        if _root.startswith(os.path.sep):
            _root = _root[len(os.path.sep):]
        if _root.startswith('.') and not include_dotfiles:
            continue
        ### ignore certain patterns
        if _found_pattern(_root):
            continue

        for filename in files:
            if filename.startswith('.') and not include_dotfiles:
                continue
            path = os.path.join(root, filename)
            if _found_pattern(path):
                continue

            _path = str(path)[len(str(_directory.parent)):]
            if _path.startswith(os.path.sep):
                _path = _path[len(os.path.sep):]
            _path = os.path.sep.join(_path.split(os.path.sep)[:-1])

            paths.append((path, _path))
    return paths

def replace_password(d : dict) -> dict:
    """
    Recursively replace passwords in a dictionary.
    """
    _d = d.copy()
    for k, v in d.items():
        if isinstance(v, dict):
            _d[k] = replace_password(v)
        elif 'password' in str(k).lower():
            _d[k] = ''.join(['*' for char in str(v)])
    return _d

def filter_keywords(
        func : Callable[[Any], Any],
        **kw : Any
    ) -> Mapping[str, Any]:
    """
    Filter out unsupported keywords.

    :param func:
        The function to inspect.
    """
    import inspect
    func_params = inspect.signature(func).parameters
    ### If the function has a **kw method, skip filtering.
    for param, _type in func_params.items():
        if '**' in str(_type):
            return kw
    func_kw = dict()
    for k, v in kw.items():
        if k in func_params:
            func_kw[k] = v
    return func_kw

def dict_from_od(od : collections.OrderedDict) -> Dict[Any, Any]:
    """
    Convert an ordered dict to a dict.
    Does not mutate the original OrderedDict.
    """
    from collections import OrderedDict
    _d = dict(od)
    for k, v in od.items():
        if isinstance(v, OrderedDict) or (
            issubclass(type(v), OrderedDict)
        ):
            _d[k] = dict_from_od(v)
    return _d

def remove_ansi(s : str) -> str:
    """
    Remove ANSI escape characters from a string.
    """
    import re
    return re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])').sub('', s)

def get_connector_labels(
        *types : str,
        search_term : str = '',
        ignore_exact_match = True,
    ) -> List[str]:
    """
    Read connector lables from config.
    """
    from meerschaum.config import get_config
    connectors = get_config('meerschaum', 'connectors')

    _types = list(types)
    if len(_types) == 0:
        _types = list(connectors.keys()) + ['plugin']

    conns = []
    for t in _types:
        if t == 'plugin':
            from meerschaum.plugins import get_data_plugins
            conns += [f'{t}:' + m.__name__.split('.')[-1] for m in get_data_plugins()]
            continue
        conns += [ f'{t}:{label}' for label in connectors.get(t, {}) if label != 'default' ]

    possibilities = [
        c for c in conns
            if c.startswith(search_term)
                and c != (
                    search_term if ignore_exact_match else ''
                )
    ]
    return sorted(possibilities)

def json_serialize_datetime(dt : datetime.datetime) -> Optional[str]:
    """
    Serialize a datetime.datetime object into JSON (ISO format string).
    """
    import datetime
    if isinstance(dt, datetime.datetime):
        return dt.isoformat() + 'Z'
    return None

def wget(
        url : str,
        dest : Optional[Union[str, pathlib.Path]] = None,
        color : bool = True,
        debug : bool = False,
        **kw : Any
    ) -> pathlib.Path:
    """
    Mimic wget with requests.
    """
    from meerschaum.utils.warnings import warn, error
    from meerschaum.utils.debug import dprint
    import os, pathlib, re, urllib.request
    if not color:
        dprint = print
    if debug:
        dprint(f"Downloading from '{url}'...")
    try:
        response = urllib.request.urlopen(url)
    except Exception as e:
        import ssl
        ssl._create_default_https_context = ssl._create_unverified_context
        try:
            response = urllib.request.urlopen(url)
        except Exception as _e:
            print(_e)
            response = None
    if response is None or response.code != 200:
        error_msg = f"Failed to download from '{url}'."
        if color:
            error(error_msg)
        else:
            print(error_msg)
            import sys
            sys.exit(1)

    d = response.headers.get('content-disposition', None)
    fname = (
        re.findall("filename=(.+)", d)[0].strip('"') if d is not None
        else url.split('/')[-1]
    )

    if dest is None:
        dest = pathlib.Path(os.path.join(os.getcwd(), fname))
    elif isinstance(dest, str):
        dest = pathlib.Path(dest)

    with open(dest, 'wb') as f:
        f.write(response.fp.read())

    if debug:
        dprint(f"Downloaded file '{dest}'.")

    return dest

def async_wrap(func):
    """
    Run a synchronous function as async.
    Found at
    https://dev.to/0xbf/turn-sync-function-to-async-python-tips-58nn
    """
    import asyncio
    from functools import wraps, partial

    @wraps(func)
    async def run(*args, loop=None, executor=None, **kwargs):
        if loop is None:
            loop = asyncio.get_event_loop()
        pfunc = partial(func, *args, **kwargs)
        return await loop.run_in_executor(executor, pfunc)
    return run 

def debug_trace(browser : bool = True):
    from meerschaum.utils.packages import attempt_import
    heartrate = attempt_import('heartrate')
    heartrate.trace(files=heartrate.files.all, browser=browser)

def items_str(
        items : List[Any],
        quotes : bool = True,
        quote_str : str = "'",
        commas : bool = True,
        comma_str : str = ',',
        and_ : bool = True,
        and_str : str = 'and',
        oxford_comma : bool = True,
        spaces : bool = True,
        space_str = ' ',
    ) -> 'str':
    """
    Return a formatted string if list items separated by commas.

    :param quotes:
        If `True`, wrap items in quotes.
        Defaults to `True`.

    :param quote_str:
        If `quotes` is `True`, prepend and append each item with this string.
        Defaults to "'" (single quote).

    :param and_:
        If `True`, include the word 'and' before the final item in the list.
        Defaults to `True`.

    :param and_str:
        If `and_` is True, insert this string where 'and' normally would in and English list. 
        Defaults to 'and'.

    :param oxford_comma:
        If `True`, include the Oxford Comma (comma before the final 'and').
        Only applies when `and_` is `True`.
        Defaults to `True`.

    :param spaces:
        If `True`, separate items with `space_str`
        Defaults to `True`.

    :param space_str:
        If `spaces` is `True`, separate items with this string.
        Defaults to ' '.
    """
    if not items:
        return ''
    
    q = quote_str if quotes else ''
    s = space_str if spaces else ''
    a = and_str if and_ else ''
    c = comma_str if commas else ''

    if len(items) == 1:
        return q + str(items[0]) + q

    if len(items) == 2:
        return q + str(items[0]) + q + s + a + s + q + str(items[1]) + q

    sep = q + c + s + q
    output = q + sep.join(str(i) for i in items[:-1]) + q
    if oxford_comma:
        output += c
    output += s + a + (s if and_ else '') + q + str(items[-1]) + q
    return output
